import functools
import logging
import os
from collections.abc import Iterator, Mapping, Sequence
from contextlib import ExitStack
from datetime import datetime
from typing import TYPE_CHECKING, Any, Callable, NamedTuple, Optional, Union, cast, overload

from typing_extensions import TypeAlias

import dagster._check as check
from dagster._annotations import beta_param, deprecated_param, public
from dagster._core.definitions.graph_definition import GraphDefinition
from dagster._core.definitions.instigation_logger import InstigationLogger
from dagster._core.definitions.job_definition import JobDefinition
from dagster._core.definitions.repository_definition import RepositoryDefinition
from dagster._core.definitions.resource_annotation import get_resource_args
from dagster._core.definitions.scoped_resources_builder import Resources, ScopedResourcesBuilder
from dagster._core.definitions.sensor_definition import (
    DagsterRunReaction,
    DefaultSensorStatus,
    RunRequest,
    SensorDefinition,
    SensorEvaluationContext,
    SensorResult,
    SensorReturnTypesUnion,
    SensorType,
    SkipReason,
    get_context_param_name,
    get_or_create_sensor_context,
    validate_and_get_resource_dict,
)
from dagster._core.definitions.target import ExecutableDefinition
from dagster._core.definitions.unresolved_asset_job_definition import UnresolvedAssetJobDefinition
from dagster._core.errors import (
    DagsterInvalidDefinitionError,
    DagsterInvariantViolationError,
    RunStatusSensorExecutionError,
    user_code_error_boundary,
)
from dagster._core.event_api import RunStatusChangeEventType, RunStatusChangeRecordsFilter
from dagster._core.events import PIPELINE_RUN_STATUS_TO_EVENT_TYPE, DagsterEvent, DagsterEventType
from dagster._core.instance import DagsterInstance
from dagster._core.storage.dagster_run import DagsterRun, DagsterRunStatus, RunsFilter
from dagster._serdes import serialize_value, whitelist_for_serdes
from dagster._serdes.errors import DeserializationError
from dagster._serdes.serdes import deserialize_value
from dagster._seven import JSONDecodeError
from dagster._time import datetime_from_timestamp, parse_time_string
from dagster._utils.error import serializable_error_info_from_exc_info
from dagster._utils.warnings import normalize_renamed_param

if TYPE_CHECKING:
    from dagster._core.definitions.resource_definition import ResourceDefinition
    from dagster._core.definitions.selector import (
        CodeLocationSelector,
        JobSelector,
        RepositorySelector,
    )

RunStatusSensorEvaluationFunction: TypeAlias = Union[
    Callable[..., SensorReturnTypesUnion],
    Callable[..., SensorReturnTypesUnion],
]
RunFailureSensorEvaluationFn: TypeAlias = Union[
    Callable[..., SensorReturnTypesUnion],
    Callable[..., SensorReturnTypesUnion],
]


def _get_run_status_sensor_fetch_limit(monitor_all_code_locations: bool) -> int:
    if monitor_all_code_locations:
        # No need to overfetch if we are going to process everything
        return _get_run_status_sensor_process_limit()

    # Otherwise, fetch more than we are planning to process, under the assumption
    # that some will be filtered out
    return int(os.getenv("DAGSTER_RUN_STATUS_SENSOR_FETCH_LIMIT", "25"))


def _get_run_status_sensor_process_limit() -> int:
    return int(os.getenv("DAGSTER_RUN_STATUS_SENSOR_PROCESS_LIMIT", "5"))


@whitelist_for_serdes(old_storage_names={"PipelineSensorCursor"})
class RunStatusSensorCursor(
    NamedTuple(
        "_RunStatusSensorCursor",
        [
            ("record_id", int),
            # deprecated arg, used as a record cursor for the run-sharded sqlite implementation to
            # filter records based on the update timestamp of the run.  When populated, the record
            # id is ignored (since it maybe run-scoped).
            ("update_timestamp", Optional[str]),
            # debug arg, used to quickly inspect the last processed timestamp from the run status
            # sensor's serialized state
            ("record_timestamp", Optional[str]),
        ],
    )
):
    def __new__(cls, record_id, update_timestamp=None, record_timestamp=None):
        return super().__new__(
            cls,
            record_id=check.int_param(record_id, "record_id"),
            update_timestamp=check.opt_str_param(update_timestamp, "update_timestamp"),
            record_timestamp=check.opt_str_param(record_timestamp, "record_timestamp"),
        )

    @staticmethod
    def is_valid(json_str: str) -> bool:
        try:
            obj = deserialize_value(json_str, RunStatusSensorCursor)
            return isinstance(obj, RunStatusSensorCursor)
        except (JSONDecodeError, DeserializationError):
            return False

    def to_json(self) -> str:
        return serialize_value(cast(NamedTuple, self))

    @staticmethod
    def from_json(json_str: str) -> "RunStatusSensorCursor":
        return deserialize_value(json_str, RunStatusSensorCursor)


class RunStatusSensorContext:
    """The ``context`` object available to a decorated function of ``run_status_sensor``."""

    def __init__(
        self,
        sensor_name,
        dagster_run,
        dagster_event,
        instance,
        context: Optional[
            SensorEvaluationContext
        ] = None,  # deprecated arg, but we need to keep it for backcompat
        resource_defs: Optional[Mapping[str, "ResourceDefinition"]] = None,
        logger: Optional[logging.Logger] = None,
        partition_key: Optional[str] = None,
        repository_def: Optional[RepositoryDefinition] = None,
        _resources: Optional[Resources] = None,
        _cm_scope_entered: bool = False,
    ) -> None:
        self._exit_stack = ExitStack()
        self._sensor_name = check.str_param(sensor_name, "sensor_name")
        self._dagster_run = check.inst_param(dagster_run, "dagster_run", DagsterRun)
        self._dagster_event = check.inst_param(dagster_event, "dagster_event", DagsterEvent)
        self._instance = check.inst_param(instance, "instance", DagsterInstance)
        self._logger: Optional[logging.Logger] = logger or (context.log if context else None)
        self._partition_key = check.opt_str_param(partition_key, "partition_key")
        self._repository_def = check.opt_inst_param(
            repository_def, "repository_def", RepositoryDefinition
        )

        # Wait to set resources unless they're accessed
        self._resource_defs = resource_defs
        self._resources = _resources
        self._cm_scope_entered = _cm_scope_entered

    def for_run_failure(self) -> "RunFailureSensorContext":
        """Converts RunStatusSensorContext to RunFailureSensorContext."""
        return RunFailureSensorContext(
            sensor_name=self._sensor_name,
            dagster_run=self._dagster_run,
            dagster_event=self._dagster_event,
            instance=self._instance,
            logger=self._logger,
            partition_key=self._partition_key,
            resource_defs=self._resource_defs,
            repository_def=self._repository_def,
            _resources=self._resources,
            _cm_scope_entered=self._cm_scope_entered,
        )

    @property
    def resource_defs(self) -> Optional[Mapping[str, "ResourceDefinition"]]:
        return self._resource_defs

    @property
    def repository_def(self) -> Optional[RepositoryDefinition]:
        """Optional[RepositoryDefinition]: The RepositoryDefinition that this sensor resides in."""
        return self._repository_def

    @property
    def resources(self) -> Resources:
        from dagster._core.definitions.scoped_resources_builder import IContainsGenerator
        from dagster._core.execution.build_resources import build_resources

        if not self._resources:
            """
            This is similar to what we do in e.g. the op context - we set up a resource
            building context manager, and immediately enter it. This is so that in cases
            where a user is not using any context-manager based resources, they don't
            need to enter this SensorEvaluationContext themselves.

            For example:

            my_sensor(build_sensor_context(resources={"my_resource": my_non_cm_resource})

            will work ok, but for a CM resource we must do

            with build_sensor_context(resources={"my_resource": my_cm_resource}) as context:
                my_sensor(context)
            """

            instance = self.instance if self._instance else None

            resources_cm = build_resources(resources=self._resource_defs or {}, instance=instance)
            self._resources = self._exit_stack.enter_context(resources_cm)

            if isinstance(self._resources, IContainsGenerator) and not self._cm_scope_entered:
                self._exit_stack.close()
                raise DagsterInvariantViolationError(
                    "At least one provided resource is a generator, but attempting to access"
                    " resources outside of context manager scope. You can use the following syntax"
                    " to open a context manager: `with build_sensor_context(...) as context:`"
                )

        return self._resources

    @public
    @property
    def sensor_name(self) -> str:
        """The name of the sensor."""
        return self._sensor_name

    @public
    @property
    def dagster_run(self) -> DagsterRun:
        """The run of the job."""
        return self._dagster_run

    @public
    @property
    def dagster_event(self) -> DagsterEvent:
        """The event associated with the job run status."""
        return self._dagster_event

    @public
    @property
    def instance(self) -> DagsterInstance:
        """The current instance."""
        return self._instance

    @public
    @property
    def log(self) -> logging.Logger:
        """The logger for the current sensor evaluation."""
        if not self._logger:
            self._logger = InstigationLogger()

        return self._logger

    @public
    @property
    def partition_key(self) -> Optional[str]:
        """Optional[str]: The partition key of the relevant run."""
        return self._partition_key

    def __enter__(self) -> "RunStatusSensorContext":
        self._cm_scope_entered = True
        return self

    def __exit__(self, *exc) -> None:
        self._exit_stack.close()
        self._logger = None

    def merge_resources(self, resources_dict: Mapping[str, Any]) -> "RunStatusSensorContext":
        """Merge the specified resources into this context.

        This method is intended to be used by the Dagster framework, and should not be called by user code.

        Args:
            resources_dict (Mapping[str, Any]): The resources to replace in the context.
        """
        check.invariant(
            self._resources is None,
            "Cannot merge resources in context that has been initialized.",
        )
        from dagster._core.execution.build_resources import wrap_resources_for_execution

        return RunStatusSensorContext(
            sensor_name=self._sensor_name,
            dagster_run=self._dagster_run,
            dagster_event=self._dagster_event,
            instance=self._instance,
            logger=self._logger,
            partition_key=self._partition_key,
            resource_defs={
                **(self._resource_defs or {}),
                **wrap_resources_for_execution(resources_dict),
            },
            repository_def=self._repository_def,
        )


class RunFailureSensorContext(RunStatusSensorContext):
    """The ``context`` object available to a decorated function of ``run_failure_sensor``.

    Args:
        sensor_name (str): the name of the sensor.
        dagster_run (DagsterRun): the failed run.
    """

    @public
    @property
    def failure_event(self) -> DagsterEvent:
        """The run failure event.

        If the run failed because of an error inside a step, get_step_failure_events will have more
        details on the step failure.
        """
        return self.dagster_event

    @public
    def get_step_failure_events(self) -> Sequence[DagsterEvent]:
        """The step failure event for each step in the run that failed.

        Examples:
            .. code-block:: python

                error_strings_by_step_key = {
                    # includes the stack trace
                    event.step_key: event.event_specific_data.error.to_string()
                    for event in context.get_step_failure_events()
                }
        """
        records = self.instance.get_records_for_run(
            run_id=self.dagster_run.run_id, of_type=DagsterEventType.STEP_FAILURE
        ).records
        return [cast(DagsterEvent, record.event_log_entry.dagster_event) for record in records]


@beta_param(param="repository_def")
def build_run_status_sensor_context(
    sensor_name: str,
    dagster_event: DagsterEvent,
    dagster_instance: DagsterInstance,
    dagster_run: DagsterRun,
    context: Optional[SensorEvaluationContext] = None,
    resources: Optional[Mapping[str, object]] = None,
    partition_key: Optional[str] = None,
    *,
    repository_def: Optional[RepositoryDefinition] = None,
) -> RunStatusSensorContext:
    """Builds run status sensor context from provided parameters.

    This function can be used to provide the context argument when directly invoking a function
    decorated with `@run_status_sensor` or `@run_failure_sensor`, such as when writing unit tests.

    Args:
        sensor_name (str): The name of the sensor the context is being constructed for.
        dagster_event (DagsterEvent): A DagsterEvent with the same event type as the one that
            triggers the run_status_sensor
        dagster_instance (DagsterInstance): The dagster instance configured for the context.
        dagster_run (DagsterRun): DagsterRun object from running a job
        resources (Optional[Mapping[str, object]]): A dictionary of resources to be made available
            to the sensor.
        repository_def (Optional[RepositoryDefinition]): The repository that the sensor belongs to.

    Examples:
        .. code-block:: python

            instance = DagsterInstance.ephemeral()
            result = my_job.execute_in_process(instance=instance)

            dagster_run = result.dagster_run
            dagster_event = result.get_job_success_event() # or get_job_failure_event()

            context = build_run_status_sensor_context(
                sensor_name="run_status_sensor_to_invoke",
                dagster_instance=instance,
                dagster_run=dagster_run,
                dagster_event=dagster_event,
            )
            run_status_sensor_to_invoke(context)
    """
    from dagster._core.execution.build_resources import wrap_resources_for_execution

    return RunStatusSensorContext(
        sensor_name=sensor_name,
        instance=dagster_instance,
        dagster_run=dagster_run,
        dagster_event=dagster_event,
        resource_defs=wrap_resources_for_execution(resources),
        logger=context.log if context else None,
        partition_key=partition_key,
        repository_def=repository_def,
    )


@overload
def run_failure_sensor(
    name: RunFailureSensorEvaluationFn,
) -> SensorDefinition: ...


@overload
def run_failure_sensor(
    name: Optional[str] = None,
    minimum_interval_seconds: Optional[int] = None,
    description: Optional[str] = None,
    monitored_jobs: Optional[
        Sequence[
            Union[
                JobDefinition,
                GraphDefinition,
                UnresolvedAssetJobDefinition,
                "RepositorySelector",
                "JobSelector",
                "CodeLocationSelector",
            ]
        ]
    ] = None,
    job_selection: Optional[
        Sequence[
            Union[
                JobDefinition,
                GraphDefinition,
                UnresolvedAssetJobDefinition,
                "RepositorySelector",
                "JobSelector",
                "CodeLocationSelector",
            ]
        ]
    ] = None,
    monitor_all_code_locations: bool = False,
    default_status: DefaultSensorStatus = DefaultSensorStatus.STOPPED,
    request_job: Optional[ExecutableDefinition] = None,
    request_jobs: Optional[Sequence[ExecutableDefinition]] = None,
    monitor_all_repositories: bool = False,
    tags: Optional[Mapping[str, str]] = None,
    metadata: Optional[Mapping[str, object]] = None,
) -> Callable[
    [RunFailureSensorEvaluationFn],
    SensorDefinition,
]: ...


@deprecated_param(
    param="job_selection",
    breaking_version="2.0",
    additional_warn_text="Use `monitored_jobs` instead.",
)
@deprecated_param(
    param="monitor_all_repositories",
    breaking_version="2.0",
    additional_warn_text="Use `monitor_all_code_locations` instead.",
)
def run_failure_sensor(
    name: Optional[Union[RunFailureSensorEvaluationFn, str]] = None,
    minimum_interval_seconds: Optional[int] = None,
    description: Optional[str] = None,
    monitored_jobs: Optional[
        Sequence[
            Union[
                JobDefinition,
                GraphDefinition,
                UnresolvedAssetJobDefinition,
                "RepositorySelector",
                "JobSelector",
                "CodeLocationSelector",
            ]
        ]
    ] = None,
    job_selection: Optional[
        Sequence[
            Union[
                JobDefinition,
                GraphDefinition,
                UnresolvedAssetJobDefinition,
                "RepositorySelector",
                "JobSelector",
                "CodeLocationSelector",
            ]
        ]
    ] = None,
    monitor_all_code_locations: Optional[bool] = None,
    default_status: DefaultSensorStatus = DefaultSensorStatus.STOPPED,
    request_job: Optional[ExecutableDefinition] = None,
    request_jobs: Optional[Sequence[ExecutableDefinition]] = None,
    monitor_all_repositories: Optional[bool] = None,
    tags: Optional[Mapping[str, str]] = None,
    metadata: Optional[Mapping[str, object]] = None,
) -> Union[
    SensorDefinition,
    Callable[
        [RunFailureSensorEvaluationFn],
        SensorDefinition,
    ],
]:
    """Creates a sensor that reacts to job failure events, where the decorated function will be
    run when a run fails.

    Takes a :py:class:`~dagster.RunFailureSensorContext`.

    Args:
        name (Optional[str]): The name of the job failure sensor. Defaults to the name of the
            decorated function.
        minimum_interval_seconds (Optional[int]): The minimum number of seconds that will elapse
            between sensor evaluations.
        description (Optional[str]): A human-readable description of the sensor.
        monitored_jobs (Optional[List[Union[JobDefinition, GraphDefinition, UnresolvedAssetJobDefinition, RepositorySelector, JobSelector, CodeLocationSelector]]]):
            The jobs in the current repository that will be monitored by this failure sensor.
            Defaults to None, which means the alert will be sent when any job in the current
            repository fails.
        monitor_all_code_locations (bool): If set to True, the sensor will monitor all runs in the
            Dagster deployment. If set to True, an error will be raised if you also specify
            monitored_jobs or job_selection. Defaults to False.
        job_selection (Optional[List[Union[JobDefinition, GraphDefinition, RepositorySelector, JobSelector, CodeLocationSelector]]]):
            (deprecated in favor of monitored_jobs) The jobs in the current repository that will be
            monitored by this failure sensor. Defaults to None, which means the alert will be sent
            when any job in the repository fails.
        default_status (DefaultSensorStatus): Whether the sensor starts as running or not. The default
            status can be overridden from the Dagster UI or via the GraphQL API.
        request_job (Optional[Union[GraphDefinition, JobDefinition, UnresolvedAssetJob]]): The job a RunRequest should
            execute if yielded from the sensor.
        request_jobs (Optional[Sequence[Union[GraphDefinition, JobDefinition, UnresolvedAssetJob]]]):
            A list of jobs to be executed if RunRequests are yielded from the sensor.
        monitor_all_repositories (bool): (deprecated in favor of monitor_all_code_locations) If set to True,
            the sensor will monitor all runs in the Dagster instance. If set to True, an error will be raised if you also specify
            monitored_jobs or job_selection. Defaults to False.
        tags (Optional[Mapping[str, str]]): A set of key-value tags that annotate the sensor and can
            be used for searching and filtering in the UI.
        metadata (Optional[Mapping[str, object]]): A set of metadata entries that annotate the
            sensor. Values will be normalized to typed `MetadataValue` objects.
    """

    def inner(
        fn: RunFailureSensorEvaluationFn,
    ) -> SensorDefinition:
        check.callable_param(fn, "fn")
        if name is None or callable(name):
            sensor_name = fn.__name__
        else:
            sensor_name = name

        jobs = monitored_jobs if monitored_jobs else job_selection
        monitor_all = normalize_renamed_param(
            monitor_all_code_locations,
            "monitor_all_code_locations",
            monitor_all_repositories,
            "monitor_all_repositories",
        )

        @run_status_sensor(
            run_status=DagsterRunStatus.FAILURE,
            name=sensor_name,
            minimum_interval_seconds=minimum_interval_seconds,
            description=description,
            monitored_jobs=jobs,
            monitor_all_code_locations=monitor_all,
            default_status=default_status,
            request_job=request_job,
            request_jobs=request_jobs,
            tags=tags,
            metadata=metadata,
        )
        @functools.wraps(fn)
        def _run_failure_sensor(*args, **kwargs) -> Any:
            args_modified = [
                arg.for_run_failure() if isinstance(arg, RunStatusSensorContext) else arg
                for arg in args
            ]
            kwargs_modified = {
                k: v.for_run_failure() if isinstance(v, RunStatusSensorContext) else v
                for k, v in kwargs.items()
            }
            return fn(*args_modified, **kwargs_modified)

        return _run_failure_sensor

    # This case is for when decorator is used bare, without arguments
    if callable(name):
        return inner(name)

    return inner


class RunStatusSensorDefinition(SensorDefinition):
    """Define a sensor that reacts to a given status of job execution, where the decorated
    function will be evaluated when a run is at the given status.

    Args:
        name (str): The name of the sensor. Defaults to the name of the decorated function.
        run_status (DagsterRunStatus): The status of a run which will be
            monitored by the sensor.
        run_status_sensor_fn (Callable[[RunStatusSensorContext], Union[SkipReason, DagsterRunReaction]]): The core
            evaluation function for the sensor. Takes a :py:class:`~dagster.RunStatusSensorContext`.
        minimum_interval_seconds (Optional[int]): The minimum number of seconds that will elapse
            between sensor evaluations.
        description (Optional[str]): A human-readable description of the sensor.
        monitored_jobs (Optional[List[Union[JobDefinition, GraphDefinition, UnresolvedAssetJobDefinition, JobSelector, RepositorySelector, CodeLocationSelector]]]):
            The jobs in the current repository that will be monitored by this sensor. Defaults to
            None, which means the alert will be sent when any job in the repository fails.
        monitor_all_code_locations (bool): If set to True, the sensor will monitor all runs in the
            Dagster deployment. If set to True, an error will be raised if you also specify
            monitored_jobs or job_selection. Defaults to False.
        default_status (DefaultSensorStatus): Whether the sensor starts as running or not. The default
            status can be overridden from the Dagster UI or via the GraphQL API.
        request_job (Optional[Union[GraphDefinition, JobDefinition]]): The job a RunRequest should
            execute if yielded from the sensor.
        tags (Optional[Mapping[str, str]]): A set of key-value tags that annotate the sensor and can
            be used for searching and filtering in the UI.
        metadata (Optional[Mapping[str, object]]): A set of metadata entries that annotate the
            sensor. Values will be normalized to typed `MetadataValue` objects.
        request_jobs (Optional[Sequence[Union[GraphDefinition, JobDefinition]]]):
            A list of jobs to be executed if RunRequests are yielded from the sensor.
    """

    def __init__(
        self,
        name: str,
        run_status: DagsterRunStatus,
        run_status_sensor_fn: RunStatusSensorEvaluationFunction,
        minimum_interval_seconds: Optional[int] = None,
        description: Optional[str] = None,
        monitored_jobs: Optional[
            Sequence[
                Union[
                    JobDefinition,
                    GraphDefinition,
                    UnresolvedAssetJobDefinition,
                    "RepositorySelector",
                    "JobSelector",
                    "CodeLocationSelector",
                ]
            ]
        ] = None,
        monitor_all_code_locations: Optional[bool] = None,
        default_status: DefaultSensorStatus = DefaultSensorStatus.STOPPED,
        request_job: Optional[ExecutableDefinition] = None,
        request_jobs: Optional[Sequence[ExecutableDefinition]] = None,
        tags: Optional[Mapping[str, str]] = None,
        metadata: Optional[Mapping[str, object]] = None,
        required_resource_keys: Optional[set[str]] = None,
    ):
        from dagster._core.definitions.selector import (
            CodeLocationSelector,
            JobSelector,
            RepositorySelector,
        )

        check.str_param(name, "name")
        check.inst_param(run_status, "run_status", DagsterRunStatus)
        check.callable_param(run_status_sensor_fn, "run_status_sensor_fn")
        check.opt_int_param(minimum_interval_seconds, "minimum_interval_seconds")
        check.opt_str_param(description, "description")
        check.opt_list_param(
            monitored_jobs,
            "monitored_jobs",
            (
                JobDefinition,
                GraphDefinition,
                UnresolvedAssetJobDefinition,
                RepositorySelector,
                JobSelector,
                CodeLocationSelector,
            ),
        )
        check.inst_param(default_status, "default_status", DefaultSensorStatus)
        monitor_all_code_locations = check.opt_bool_param(
            monitor_all_code_locations, "monitor_all_code_locations", default=False
        )

        resource_arg_names: set[str] = {arg.name for arg in get_resource_args(run_status_sensor_fn)}

        combined_required_resource_keys = (
            check.opt_set_param(required_resource_keys, "required_resource_keys", of_type=str)
            | resource_arg_names
        )

        # coerce CodeLocationSelectors to RepositorySelectors with repo name "__repository__"
        monitored_jobs = [
            job.to_repository_selector() if isinstance(job, CodeLocationSelector) else job
            for job in (monitored_jobs or [])
        ]

        self._run_status_sensor_fn = check.callable_param(
            run_status_sensor_fn, "run_status_sensor_fn"
        )
        event_type = PIPELINE_RUN_STATUS_TO_EVENT_TYPE[run_status]

        # split monitored_jobs into external repos, external jobs, and jobs in the current repo
        other_repos = (
            [x for x in monitored_jobs if isinstance(x, RepositorySelector)]
            if monitored_jobs
            else []
        )

        other_repo_jobs = (
            [x for x in monitored_jobs if isinstance(x, JobSelector)] if monitored_jobs else []
        )

        current_repo_jobs = (
            [x for x in monitored_jobs if not isinstance(x, (JobSelector, RepositorySelector))]
            if monitored_jobs
            else []
        )

        def _wrapped_fn(
            context: SensorEvaluationContext,
        ) -> Iterator[Union[RunRequest, SkipReason, DagsterRunReaction, SensorResult]]:
            # initiate the cursor to (most recent event id, current timestamp) when:
            # * it's the first time starting the sensor
            # * or, the cursor isn't in valid format (backcompt)
            if context.cursor is None or not RunStatusSensorCursor.is_valid(context.cursor):
                most_recent_event_records = context.instance.fetch_run_status_changes(
                    records_filter=event_type, limit=1
                ).records
                most_recent_event_id = (
                    most_recent_event_records[0].storage_id
                    if len(most_recent_event_records) == 1
                    else -1
                )
                record_timestamp = (
                    datetime_from_timestamp(most_recent_event_records[0].timestamp).isoformat()
                    if len(most_recent_event_records) == 1
                    else None
                )

                new_cursor = RunStatusSensorCursor(
                    record_id=most_recent_event_id, record_timestamp=record_timestamp
                )
                context.update_cursor(new_cursor.to_json())
                yield SkipReason(f"Initiating {name}. Set cursor to {new_cursor}")
                return

            sensor_cursor = RunStatusSensorCursor.from_json(context.cursor)

            process_limit = _get_run_status_sensor_process_limit()

            fetch_limit = _get_run_status_sensor_fetch_limit(
                monitor_all_code_locations=cast(bool, monitor_all_code_locations)
            )

            # Fetch events after the cursor id
            # * we move the cursor forward to the latest visited event's id to avoid revisits
            # * when the daemon is down, bc we persist the cursor info, we can go back to where we
            #   left and backfill alerts for the qualified events during the downtime
            if sensor_cursor.update_timestamp and context.instance.event_log_storage.is_run_sharded:
                # The run status sensor cursor has the timestamp set... and the event log storage
                # is run sharded.  We need to query the index shard by timestamp instead of by
                # record id (which is reindexed relative to some run sharded query).  When we update
                # the cursor, we should omit the timestamp, since this API only queries the global
                # index shard instead of the run shard.
                event_records = context.instance.fetch_run_status_changes(
                    records_filter=RunStatusChangeRecordsFilter(
                        event_type=cast(RunStatusChangeEventType, event_type),
                        after_timestamp=cast(
                            datetime, parse_time_string(sensor_cursor.update_timestamp)
                        ).timestamp(),
                    ),
                    ascending=True,
                    limit=fetch_limit,
                ).records
            elif (
                context.instance.event_log_storage.supports_run_status_change_job_name_filter
                and monitored_jobs
                and all(
                    [
                        not isinstance(monitored, (RepositorySelector, CodeLocationSelector))
                        for monitored in monitored_jobs
                    ]
                )
            ):
                # the event log storage supports run status change selectors... we should construct
                # the appropriate job selectors so that we can filter the events by jobs in the
                # storage layer instead of in memory.  This should improve throughput since we will
                # avoid fetching events that we will filter out later on.
                job_names = _job_names_for_monitored(
                    cast(
                        Sequence[
                            Union[
                                JobDefinition,
                                GraphDefinition,
                                UnresolvedAssetJobDefinition,
                                "JobSelector",
                            ]
                        ],
                        monitored_jobs,
                    )
                )
                event_records = context.instance.fetch_run_status_changes(
                    records_filter=RunStatusChangeRecordsFilter(
                        event_type=cast(RunStatusChangeEventType, event_type),
                        after_storage_id=sensor_cursor.record_id,
                        job_names=job_names,
                    ),
                    ascending=True,
                    limit=fetch_limit,
                ).records
            else:
                # the cursor storage id is globally unique, either because the event log storage is
                # not run sharded or because the cursor was set from an event returned from the
                # index shard. When we update the cursor, we should omit the timestamp, since this
                # API only queries the global index shard instead of the run shard.
                event_records = context.instance.fetch_run_status_changes(
                    records_filter=RunStatusChangeRecordsFilter(
                        event_type=cast(RunStatusChangeEventType, event_type),
                        after_storage_id=sensor_cursor.record_id,
                    ),
                    ascending=True,
                    limit=fetch_limit,
                ).records

            run_ids_to_fetch = list(
                set(event_record.event_log_entry.run_id for event_record in event_records)
            )

            run_records = (
                {
                    record.dagster_run.run_id: record
                    for record in context.instance.get_run_records(
                        filters=RunsFilter(run_ids=run_ids_to_fetch)
                    )
                }
                if run_ids_to_fetch
                else {}
            )

            num_processed_runs = 0
            for event_record in event_records:
                event_log_entry = event_record.event_log_entry
                storage_id = event_record.storage_id
                record_timestamp = datetime_from_timestamp(event_record.timestamp).isoformat()

                # skip if we couldn't find the right run
                if event_log_entry.run_id not in run_records:
                    context.update_cursor(
                        RunStatusSensorCursor(
                            record_id=storage_id, record_timestamp=record_timestamp
                        ).to_json()
                    )
                    continue

                dagster_run = run_records[event_log_entry.run_id].dagster_run
                job_match = False

                # if monitor_all_code_locations is provided, then we want to run the sensor for all jobs in all code locations
                if monitor_all_code_locations:
                    job_match = True

                sensor_location_name = (
                    context.code_location_origin.location_name
                    if context.code_location_origin
                    else None
                )
                sensor_repo_name = context.repository_name
                dagster_run_location_name = (
                    dagster_run.remote_job_origin.repository_origin.code_location_origin.location_name
                    if dagster_run.remote_job_origin
                    else None
                )
                dagster_run_repo_name = (
                    dagster_run.remote_job_origin.repository_origin.repository_name
                    if dagster_run.remote_job_origin
                    else None
                )

                # check if the run is in the current repository and (if provided) one of jobs specified in monitored_jobs
                # note: this also matches when both are None for testing uses cases
                if (
                    not job_match
                    and (sensor_location_name == dagster_run_location_name)
                    and (sensor_repo_name == dagster_run_repo_name)
                ):
                    if monitored_jobs:
                        if dagster_run.job_name in map(lambda x: x.name, current_repo_jobs):
                            job_match = True
                    else:
                        job_match = True

                if (
                    not job_match
                    and
                    # the job has a repository (not manually executed)
                    dagster_run.remote_job_origin
                ):
                    # check if the run is one of the jobs specified by JobSelector or RepositorySelector (ie in another repo)
                    # make a JobSelector for the run in question
                    remote_repository_origin = dagster_run.remote_job_origin.repository_origin
                    run_job_selector = JobSelector(
                        location_name=remote_repository_origin.code_location_origin.location_name,
                        repository_name=remote_repository_origin.repository_name,
                        job_name=dagster_run.job_name,
                    )
                    if run_job_selector in other_repo_jobs:
                        job_match = True

                    # make a RepositorySelector for the run in question
                    run_repo_selector = RepositorySelector(
                        location_name=remote_repository_origin.code_location_origin.location_name,
                        repository_name=remote_repository_origin.repository_name,
                    )
                    if run_repo_selector in other_repos:
                        job_match = True

                if not job_match:
                    # the run in question doesn't match any of the criteria for we advance the cursor and move on
                    context.update_cursor(
                        RunStatusSensorCursor(
                            record_id=storage_id, record_timestamp=record_timestamp
                        ).to_json()
                    )
                    continue

                # Stop processing runs once you reach a matching job but have exceeded the limit
                # (It's fine to keep advancing the cursor for runs that do not match)
                if num_processed_runs >= process_limit:
                    break

                num_processed_runs = num_processed_runs + 1

                serializable_error = None

                resource_args_populated = validate_and_get_resource_dict(
                    context.resources, name, resource_arg_names
                )

                try:
                    with (
                        RunStatusSensorContext(
                            sensor_name=name,
                            dagster_run=dagster_run,
                            dagster_event=event_log_entry.dagster_event,
                            instance=context.instance,
                            resource_defs=context.resource_defs,
                            logger=context.log,
                            partition_key=dagster_run.tags.get("dagster/partition"),
                            repository_def=context.repository_def,
                        ) as sensor_context,
                        user_code_error_boundary(
                            RunStatusSensorExecutionError,
                            lambda: f'Error occurred during the execution sensor "{name}".',
                        ),
                    ):
                        context_param_name = get_context_param_name(run_status_sensor_fn)
                        context_param = (
                            {context_param_name: sensor_context} if context_param_name else {}
                        )

                        sensor_return = run_status_sensor_fn(
                            **context_param,
                            **resource_args_populated,
                        )

                        if sensor_return is not None:
                            context.update_cursor(
                                RunStatusSensorCursor(
                                    record_id=storage_id,
                                    record_timestamp=record_timestamp,
                                ).to_json()
                            )

                            if isinstance(sensor_return, SensorResult):
                                if sensor_return.cursor:
                                    raise DagsterInvariantViolationError(
                                        f"Error in run status sensor {name}: Sensor returned a"
                                        " SensorResult with a cursor value. The cursor is managed"
                                        " by the sensor and should not be modified by a user."
                                    )
                                yield sensor_return
                            elif isinstance(
                                sensor_return,
                                (RunRequest, SkipReason, DagsterRunReaction),
                            ):
                                yield sensor_return
                            else:
                                yield from sensor_return
                            return
                except RunStatusSensorExecutionError as run_status_sensor_execution_error:
                    # When the user code errors, we report error to the sensor tick not the original run.
                    serializable_error = serializable_error_info_from_exc_info(
                        run_status_sensor_execution_error.original_exc_info
                    )

                context.update_cursor(
                    RunStatusSensorCursor(
                        record_id=storage_id, record_timestamp=record_timestamp
                    ).to_json()
                )

                # Yield DagsterRunReaction to indicate the execution success/failure.
                # The sensor machinery would
                # * report back to the original run if success
                # * update cursor and job state
                yield DagsterRunReaction(
                    dagster_run=dagster_run,
                    run_status=run_status,
                    error=serializable_error,
                )

        super().__init__(
            name=name,
            evaluation_fn=_wrapped_fn,
            minimum_interval_seconds=minimum_interval_seconds,
            description=description,
            default_status=default_status,
            job=request_job,
            jobs=request_jobs,
            required_resource_keys=combined_required_resource_keys,
            tags=tags,
            metadata=metadata,
        )

    def __call__(self, *args, **kwargs) -> SensorReturnTypesUnion:
        context_param_name = get_context_param_name(self._run_status_sensor_fn)
        context = get_or_create_sensor_context(
            self._run_status_sensor_fn,
            *args,
            context_type=RunStatusSensorContext,
            **kwargs,
        )
        context_param = {context_param_name: context} if context_param_name and context else {}

        resources = validate_and_get_resource_dict(
            context.resources if context else ScopedResourcesBuilder.build_empty(),
            self._name,
            self._required_resource_keys,
        )
        return self._run_status_sensor_fn(**context_param, **resources)

    @property
    def sensor_type(self) -> SensorType:
        return SensorType.RUN_STATUS


@deprecated_param(
    param="job_selection",
    breaking_version="2.0",
    additional_warn_text="Use `monitored_jobs` instead.",
)
@deprecated_param(
    param="monitor_all_repositories",
    breaking_version="2.0",
    additional_warn_text="Use `monitor_all_code_locations` instead.",
)
def run_status_sensor(
    run_status: DagsterRunStatus,
    name: Optional[str] = None,
    minimum_interval_seconds: Optional[int] = None,
    description: Optional[str] = None,
    monitored_jobs: Optional[
        Sequence[
            Union[
                JobDefinition,
                GraphDefinition,
                UnresolvedAssetJobDefinition,
                "RepositorySelector",
                "JobSelector",
                "CodeLocationSelector",
            ]
        ]
    ] = None,
    job_selection: Optional[
        Sequence[
            Union[
                JobDefinition,
                GraphDefinition,
                UnresolvedAssetJobDefinition,
                "RepositorySelector",
                "JobSelector",
                "CodeLocationSelector",
            ]
        ]
    ] = None,
    monitor_all_code_locations: Optional[bool] = None,
    default_status: DefaultSensorStatus = DefaultSensorStatus.STOPPED,
    request_job: Optional[ExecutableDefinition] = None,
    request_jobs: Optional[Sequence[ExecutableDefinition]] = None,
    monitor_all_repositories: Optional[bool] = None,
    tags: Optional[Mapping[str, str]] = None,
    metadata: Optional[Mapping[str, object]] = None,
) -> Callable[
    [RunStatusSensorEvaluationFunction],
    RunStatusSensorDefinition,
]:
    """Creates a sensor that reacts to a given status of job execution, where the decorated
    function will be run when a job is at the given status.

    Takes a :py:class:`~dagster.RunStatusSensorContext`.

    Args:
        run_status (DagsterRunStatus): The status of run execution which will be
            monitored by the sensor.
        name (Optional[str]): The name of the sensor. Defaults to the name of the decorated function.
        minimum_interval_seconds (Optional[int]): The minimum number of seconds that will elapse
            between sensor evaluations.
        description (Optional[str]): A human-readable description of the sensor.
        monitored_jobs (Optional[List[Union[JobDefinition, GraphDefinition, UnresolvedAssetJobDefinition, RepositorySelector, JobSelector, CodeLocationSelector]]]):
            Jobs in the current code locations that will be monitored by this sensor. Defaults to None, which means the alert will
            be sent when any job in the code location matches the requested run_status. Jobs in external repositories can be monitored by using
            RepositorySelector or JobSelector.
        monitor_all_code_locations (Optional[bool]): If set to True, the sensor will monitor all runs in the Dagster deployment.
            If set to True, an error will be raised if you also specify monitored_jobs or job_selection.
            Defaults to False.
        job_selection (Optional[List[Union[JobDefinition, GraphDefinition, RepositorySelector, JobSelector, CodeLocationSelector]]]):
            (deprecated in favor of monitored_jobs) Jobs in the current code location that will be
            monitored by this sensor. Defaults to None, which means the alert will be sent when
            any job in the code location matches the requested run_status.
        default_status (DefaultSensorStatus): Whether the sensor starts as running or not. The default
            status can be overridden from the Dagster UI or via the GraphQL API.
        request_job (Optional[Union[GraphDefinition, JobDefinition, UnresolvedAssetJobDefinition]]): The job that should be
            executed if a RunRequest is yielded from the sensor.
        request_jobs (Optional[Sequence[Union[GraphDefinition, JobDefinition, UnresolvedAssetJobDefinition]]]):
            A list of jobs to be executed if RunRequests are yielded from the sensor.
        monitor_all_repositories (Optional[bool]): (deprecated in favor of monitor_all_code_locations) If set to True, the sensor will monitor all runs in the Dagster instance.
            If set to True, an error will be raised if you also specify monitored_jobs or job_selection.
            Defaults to False.
        tags (Optional[Mapping[str, str]]): A set of key-value tags that annotate the sensor and can
            be used for searching and filtering in the UI.
        metadata (Optional[Mapping[str, object]]): A set of metadata entries that annotate the
            sensor. Values will be normalized to typed `MetadataValue` objects.
    """

    def inner(
        fn: RunStatusSensorEvaluationFunction,
    ) -> RunStatusSensorDefinition:
        check.callable_param(fn, "fn")
        sensor_name = name or fn.__name__

        jobs = monitored_jobs if monitored_jobs else job_selection
        monitor_all = normalize_renamed_param(
            monitor_all_code_locations,
            "monitor_all_code_locations",
            monitor_all_repositories,
            "monitor_all_repositories",
        )

        if jobs and monitor_all:
            raise DagsterInvalidDefinitionError(
                f"Cannot specify both {'monitor_all_code_locations' if monitor_all_code_locations else 'monitor_all_repositories'} and"
                f" {'monitored_jobs' if monitored_jobs else 'job_selection'}."
            )

        return RunStatusSensorDefinition(
            name=sensor_name,
            run_status=run_status,
            run_status_sensor_fn=fn,
            minimum_interval_seconds=minimum_interval_seconds,
            description=description,
            monitored_jobs=jobs,
            monitor_all_code_locations=monitor_all,
            default_status=default_status,
            request_job=request_job,
            request_jobs=request_jobs,
            tags=tags,
            metadata=metadata,
        )

    return inner


def _job_names_for_monitored(
    monitored: Sequence[
        Union[
            JobDefinition,
            GraphDefinition,
            UnresolvedAssetJobDefinition,
            "JobSelector",
        ]
    ],
) -> Sequence[str]:
    from dagster._core.definitions.selector import JobSelector

    job_names = []
    for m in monitored:
        if isinstance(m, JobSelector):
            job_names.append(m.job_name)
        else:
            job_names.append(m.name)
    return job_names



from abc import ABC, abstractmethod
from collections.abc import Iterator, Mapping, Sequence
from contextlib import contextmanager
from typing import Any, Generic, NamedTuple, Optional, TypeVar, Union, cast

import dagster._check as check
from dagster._check import CheckError
from dagster._core.definitions.metadata import RawMetadataValue
from dagster._core.definitions.metadata.metadata_set import TableMetadataSet
from dagster._core.definitions.multi_dimensional_partitions import (
    MultiPartitionKey,
    MultiPartitionsDefinition,
)
from dagster._core.definitions.time_window_partitions import (
    TimeWindow,
    TimeWindowPartitionsDefinition,
)
from dagster._core.errors import DagsterInvalidMetadata, DagsterInvariantViolationError
from dagster._core.execution.context.input import InputContext
from dagster._core.execution.context.output import OutputContext
from dagster._core.storage.io_manager import IOManager

T = TypeVar("T")


class TablePartitionDimension(NamedTuple):
    partition_expr: str
    partitions: Union[TimeWindow, Sequence[str]]


class TableSlice(NamedTuple):
    table: str
    schema: str
    database: Optional[str] = None
    columns: Optional[Sequence[str]] = None
    partition_dimensions: Optional[Sequence[TablePartitionDimension]] = None


class DbTypeHandler(ABC, Generic[T]):
    @abstractmethod
    def handle_output(
        self, context: OutputContext, table_slice: TableSlice, obj: T, connection
    ) -> Optional[Mapping[str, RawMetadataValue]]:
        """Stores the given object at the given table in the given schema."""

    @abstractmethod
    def load_input(self, context: InputContext, table_slice: TableSlice, connection) -> T:
        """Loads the contents of the given table in the given schema."""

    @property
    @abstractmethod
    def supported_types(self) -> Sequence[type[object]]:
        pass


class DbClient(Generic[T]):
    @staticmethod
    @abstractmethod
    def delete_table_slice(
        context: OutputContext, table_slice: TableSlice, connection: T
    ) -> None: ...

    @staticmethod
    @abstractmethod
    def get_select_statement(table_slice: TableSlice) -> str: ...

    @staticmethod
    def get_table_name(table_slice: TableSlice) -> Optional[str]:
        """Returns a string which is set as the dagster/table_name metadata value for an
        emitted asset. This value should be the fully qualified name of the table, including the
        schema and database, if applicable.
        """
        if not table_slice.database:
            return f"{table_slice.schema}.{table_slice.table}"

        return f"{table_slice.database}.{table_slice.schema}.{table_slice.table}"

    @staticmethod
    @abstractmethod
    def ensure_schema_exists(
        context: OutputContext, table_slice: TableSlice, connection: T
    ) -> None: ...

    @staticmethod
    @abstractmethod
    @contextmanager
    def connect(
        context: Union[OutputContext, InputContext], table_slice: TableSlice
    ) -> Iterator[T]: ...


class DbIOManager(IOManager):
    def __init__(
        self,
        *,
        type_handlers: Sequence[DbTypeHandler],
        db_client: DbClient,
        database: str,
        schema: Optional[str] = None,
        io_manager_name: Optional[str] = None,
        default_load_type: Optional[type] = None,
    ):
        self._handlers_by_type: dict[type[Any], DbTypeHandler] = {}
        self._io_manager_name = io_manager_name or self.__class__.__name__
        for type_handler in type_handlers:
            for handled_type in type_handler.supported_types:
                check.invariant(
                    handled_type not in self._handlers_by_type,
                    f"{self._io_manager_name} provided with two handlers for the same type. "
                    f"Type: '{handled_type}'. Handler classes: '{type(type_handler)}' and "
                    f"'{type(self._handlers_by_type.get(handled_type))}'.",
                )

                self._handlers_by_type[handled_type] = type_handler
        self._db_client = db_client
        self._database = database
        self._schema = schema
        if (
            default_load_type is None
            and len(type_handlers) == 1
            and len(type_handlers[0].supported_types) == 1
        ):
            self._default_load_type = type_handlers[0].supported_types[0]
        else:
            self._default_load_type = default_load_type

    def handle_output(self, context: OutputContext, obj: object) -> None:
        # If the output type is set to Nothing, handle_output will not be
        # called. We still need to raise an error when the return value
        # is None, but the typing type is not Nothing
        if obj is None:
            raise DagsterInvariantViolationError(
                "Unexpected 'None' output value. If a 'None' value is intentional, set the output"
                " type to None by adding return type annotation '-> None'.",
            )

        obj_type = type(obj)
        self._check_supported_type(obj_type)

        table_slice = self._get_table_slice(context, context)

        with self._db_client.connect(context, table_slice) as conn:
            self._db_client.ensure_schema_exists(context, table_slice, conn)
            self._db_client.delete_table_slice(context, table_slice, conn)

            handler = self._resolve_handler(obj_type)
            handler_metadata = handler.handle_output(context, table_slice, obj, conn)

        context.add_output_metadata(
            {
                **(handler_metadata or {}),
                "Query": self._db_client.get_select_statement(table_slice),
            }
        )

        # Try to attach relation identifier metadata to the output asset, but
        # don't fail if it errors because the user has already attached it.
        try:
            context.add_output_metadata(
                dict(TableMetadataSet(table_name=self._db_client.get_table_name(table_slice)))
            )
        except DagsterInvalidMetadata:
            pass

    def load_input(self, context: InputContext) -> object:
        obj_type = context.dagster_type.typing_type
        if obj_type is Any and self._default_load_type is not None:
            load_type = self._default_load_type
        else:
            load_type = obj_type

        self._check_supported_type(load_type)

        table_slice = self._get_table_slice(context, cast(OutputContext, context.upstream_output))

        with self._db_client.connect(context, table_slice) as conn:
            return self._resolve_handler(load_type).load_input(context, table_slice, conn)  # type: ignore  # (pyright bug)

    def _resolve_handler(self, obj_type: type) -> DbTypeHandler:
        return next(
            handler
            for type_, handler in self._handlers_by_type.items()
            if issubclass(obj_type, type_)
        )

    def _get_table_slice(
        self, context: Union[OutputContext, InputContext], output_context: OutputContext
    ) -> TableSlice:
        output_context_metadata = output_context.definition_metadata or {}

        schema: str
        table: str
        partition_dimensions: list[TablePartitionDimension] = []
        if context.has_asset_key:
            asset_key_path = context.asset_key.path
            table = asset_key_path[-1]
            # schema order of precedence: metadata, I/O manager 'schema' config, key_prefix
            if output_context_metadata.get("schema"):
                schema = cast(str, output_context_metadata["schema"])
            elif self._schema:
                schema = self._schema
            elif len(asset_key_path) > 1:
                schema = asset_key_path[-2]
            else:
                schema = "public"

            if context.has_asset_partitions:
                partition_expr = output_context_metadata.get("partition_expr")
                if partition_expr is None:
                    raise ValueError(
                        f"Asset '{context.asset_key}' has partitions, but no 'partition_expr'"
                        " metadata value, so we don't know what column it's partitioned on. To"
                        " specify a column, set this metadata value. E.g."
                        ' @asset(metadata={"partition_expr": "your_partition_column"}).'
                    )

                if isinstance(context.asset_partitions_def, MultiPartitionsDefinition):
                    multi_partition_key_mapping = cast(
                        MultiPartitionKey, context.asset_partition_key
                    ).keys_by_dimension
                    for part in context.asset_partitions_def.partitions_defs:
                        partition_key = multi_partition_key_mapping[part.name]
                        if isinstance(part.partitions_def, TimeWindowPartitionsDefinition):
                            partitions = part.partitions_def.time_window_for_partition_key(
                                partition_key
                            )
                        else:
                            partitions = [partition_key]

                        partition_expr_str = cast(Mapping[str, str], partition_expr).get(part.name)
                        if partition_expr is None:
                            raise ValueError(
                                f"Asset '{context.asset_key}' has partition {part.name}, but the"
                                f" 'partition_expr' metadata does not contain a {part.name} entry,"
                                " so we don't know what column to filter it on. Specify which"
                                " column of the database contains data for the"
                                f" {part.name} partition."
                            )
                        partition_dimensions.append(
                            TablePartitionDimension(
                                partition_expr=cast(str, partition_expr_str), partitions=partitions
                            )
                        )
                elif isinstance(context.asset_partitions_def, TimeWindowPartitionsDefinition):
                    partition_dimensions.append(
                        TablePartitionDimension(
                            partition_expr=cast(str, partition_expr),
                            partitions=(
                                context.asset_partitions_time_window
                                if context.asset_partition_keys
                                else []
                            ),
                        )
                    )
                else:
                    partition_dimensions.append(
                        TablePartitionDimension(
                            partition_expr=cast(str, partition_expr),
                            partitions=context.asset_partition_keys,
                        )
                    )
        else:
            table = output_context.name
            if output_context_metadata.get("schema"):
                schema = cast(str, output_context_metadata["schema"])
            elif self._schema:
                schema = self._schema
            else:
                schema = "public"

        return TableSlice(
            table=table,
            schema=schema,
            database=self._database,
            partition_dimensions=partition_dimensions,
            columns=(context.definition_metadata or {}).get("columns"),
        )

    def _check_supported_type(self, obj_type):
        if not issubclass(obj_type, tuple(self._handlers_by_type.keys())):
            msg = (
                f"{self._io_manager_name} does not have a handler for type '{obj_type}'. Has"
                " handlers for types"
                f" '{', '.join([str(handler_type) for handler_type in self._handlers_by_type.keys()])}'."
            )

            if obj_type is Any:
                type_hints = " or ".join(
                    [str(handler_type) for handler_type in self._handlers_by_type.keys()]
                )
                msg += f" Please add {type_hints} type hints to your assets and ops."
            else:
                msg += (
                    f" Please build the {self._io_manager_name} with an type handler for type"
                    f" '{obj_type}', so the {self._io_manager_name} can correctly handle the"
                    " output."
                )

            raise CheckError(msg)



"""add_run_tags_run_id_index

Revision ID: 284a732df317
Revises: 46b412388816
Create Date: 2024-07-10 09:35:20.215174

"""

from alembic import op
from dagster._core.storage.migration.utils import has_index, has_table

# revision identifiers, used by Alembic.
revision = "284a732df317"
down_revision = "46b412388816"
branch_labels = None
depends_on = None


def upgrade():
    if not has_table("run_tags"):
        return

    if not has_index("run_tags", "idx_run_tags_run_idx"):
        op.create_index(
            "idx_run_tags_run_idx",
            "run_tags",
            ["run_id", "id"],
            unique=False,
            postgresql_concurrently=True,
            mysql_length={"run_id": 255},
        )


def downgrade():
    if not has_table("run_tags"):
        return

    if has_index("run_tags", "idx_run_tags_run_idx"):
        op.drop_index(
            "idx_run_tags_run_idx",
            "run_tags",
            postgresql_concurrently=True,
        )



"""This file contains the typing api that should exist in python in
order to do metaprogramming and reflection on the built-in typing module.
"""

import typing

from typing_extensions import get_args, get_origin

import dagster._check as check

try:
    # this type only exists in python 3.10+
    from types import UnionType  # type: ignore
except ImportError:
    UnionType = typing.Union


def is_closed_python_optional_type(annotation) -> bool:
    """Returns true if the annotation signifies an Optional type
    that is closed over a non-None T.

    In particular, this can be:
    - Optional[T]
    - Union[T, None]
    - Union[None, T]
    - T | None (in Python 3.10+)
    - None | T (in Python 3.10+).

    """
    # Optional[T] is equivalent to Union[T, None], see
    # https://docs.python.org/3/library/typing.html#typing.Optional
    # A user could also specify a Union themselves
    if get_origin(annotation) == typing.Union:
        return len(get_args(annotation)) == 2 and type(None) in get_args(annotation)

    # The Python 3.10 pipe syntax evaluates to a UnionType
    # rather than a Union, so we need to check for that as well
    # UnionType values are equivalent to Unions, e.g. str | None == Union[str, None]
    # but the types themselves are not, e.g. type(str | None) != type(Union[str, None])
    if get_origin(annotation) == UnionType:
        return len(get_args(annotation)) == 2 and type(None) in get_args(annotation)

    return False


def is_python_dict_type(ttype):
    origin = get_origin(ttype)
    return ttype is dict or origin is dict


def is_closed_python_list_type(ttype):
    origin = get_origin(ttype)
    args = get_args(ttype)

    return origin is list and args != ()


def is_closed_python_dict_type(ttype):
    """A "closed" generic type has all of its type parameters parameterized
    by other closed or concrete types.

    e.g.

    Returns true for typing.Dict[int, str] but not for typing.Dict.

    Tests document current behavior (not recursive) -- i.e., typing.Dict[str, Dict] returns True.
    """
    origin = get_origin(ttype)
    args = get_args(ttype)

    return origin is dict and args != ()


def is_closed_python_tuple_type(ttype):
    """A "closed" generic type has all of its type parameters parameterized
    by other closed or concrete types.

    e.g.

    Returns true for Tuple[int] or Tuple[str, int] but false for Tuple or tuple
    """
    origin = get_origin(ttype)
    args = get_args(ttype)

    return origin is tuple and args != ()


def is_closed_python_set_type(ttype):
    """A "closed" generic type has all of its type parameters parameterized
    by other closed or concrete types.

    e.g.

    Returns true for Set[string] but false for Set or set
    """
    origin = get_origin(ttype)
    args = get_args(ttype)

    return origin is set and args != ()


def get_optional_inner_type(ttype):
    check.invariant(
        is_closed_python_optional_type(ttype), "type must pass is_closed_python_optional_type check"
    )

    return get_args(ttype)[0]


def get_list_inner_type(ttype):
    check.param_invariant(is_closed_python_list_type(ttype), "ttype")
    return get_args(ttype)[0]


def get_set_inner_type(ttype):
    check.param_invariant(is_closed_python_set_type(ttype), "ttype")
    return get_args(ttype)[0]


def get_tuple_type_params(ttype):
    check.param_invariant(is_closed_python_tuple_type(ttype), "ttype")
    return get_args(ttype)


def get_dict_key_value_types(ttype):
    check.param_invariant(is_closed_python_dict_type(ttype), "ttype")
    return get_args(ttype)


def is_typing_type(ttype):
    return (
        is_closed_python_dict_type(ttype)
        or is_closed_python_optional_type(ttype)
        or is_closed_python_set_type(ttype)
        or is_closed_python_tuple_type(ttype)
        or is_closed_python_list_type(ttype)
        or ttype is typing.Tuple  # noqa: UP006
        or ttype is typing.Set  # noqa: UP006
        or ttype is typing.Dict  # noqa: UP006
        or ttype is typing.List  # noqa: UP006
    )


def flatten_unions(ttype: type) -> typing.AbstractSet[type]:
    """Accepts a type that may be a Union of other types, and returns those other types.
    In addition to explicit Union annotations, works for Optional, which is represented as
    Union[T, None] under the covers.

    E.g. Optional[Union[str, Union[int, float]]] would result in (str, int, float, type(None))
    """
    return set(_flatten_unions_inner(ttype))


def _flatten_unions_inner(ttype: type) -> typing.Iterable[type]:
    if get_origin(ttype) is typing.Union:
        for arg in get_args(ttype):
            yield from flatten_unions(arg)
    else:
        yield ttype



import logging
import time
from collections.abc import Iterable, Mapping, Sequence
from typing import Any, Callable, Optional, Union, cast

import dagster._check as check
from dagster._core.definitions.partition import PartitionsDefinition
from dagster._core.definitions.partition_key_range import PartitionKeyRange
from dagster._core.definitions.selector import JobSubsetSelector
from dagster._core.errors import DagsterBackfillFailedError, DagsterInvariantViolationError
from dagster._core.execution.backfill import (
    BulkActionStatus,
    PartitionBackfill,
    cancel_backfill_runs_and_cancellation_complete,
)
from dagster._core.execution.plan.resume_retry import ReexecutionStrategy
from dagster._core.execution.plan.state import KnownExecutionState
from dagster._core.instance import DagsterInstance
from dagster._core.remote_representation import CodeLocation, RemoteJob, RemotePartitionSet
from dagster._core.remote_representation.external_data import PartitionSetExecutionParamSnap
from dagster._core.remote_representation.origin import RemotePartitionSetOrigin
from dagster._core.storage.dagster_run import (
    NOT_FINISHED_STATUSES,
    DagsterRun,
    DagsterRunStatus,
    RunsFilter,
)
from dagster._core.storage.tags import (
    ASSET_PARTITION_RANGE_END_TAG,
    ASSET_PARTITION_RANGE_START_TAG,
    PARENT_RUN_ID_TAG,
    PARTITION_NAME_TAG,
    PARTITION_SET_TAG,
    ROOT_RUN_ID_TAG,
)
from dagster._core.telemetry import BACKFILL_RUN_CREATED, hash_name, log_action
from dagster._core.utils import make_new_run_id
from dagster._core.workspace.context import BaseWorkspaceRequestContext, IWorkspaceProcessContext
from dagster._time import get_current_timestamp
from dagster._utils import check_for_debug_crash
from dagster._utils.error import SerializableErrorInfo
from dagster._utils.merger import merge_dicts

# out of abundance of caution, sleep at checkpoints in case we are pinning CPU by submitting lots
# of jobs all at once
CHECKPOINT_INTERVAL = 1
CHECKPOINT_COUNT = 25


def execute_job_backfill_iteration(
    backfill: PartitionBackfill,
    logger: logging.Logger,
    workspace_process_context: IWorkspaceProcessContext,
    debug_crash_flags: Optional[Mapping[str, int]],
    instance: DagsterInstance,
) -> Iterable[Optional[SerializableErrorInfo]]:
    if not backfill.last_submitted_partition_name:
        logger.info(f"Starting job backfill for {backfill.backfill_id}")
    else:
        logger.info(
            f"Resuming job backfill for {backfill.backfill_id} from"
            f" {backfill.last_submitted_partition_name}"
        )

    partition_set = _get_partition_set(workspace_process_context, backfill)

    # refetch in case the backfill status has changed
    backfill = cast(PartitionBackfill, instance.get_backfill(backfill.backfill_id))
    if backfill.status == BulkActionStatus.CANCELING:
        for all_runs_canceled in cancel_backfill_runs_and_cancellation_complete(
            instance=instance, backfill_id=backfill.backfill_id
        ):
            yield None

        if not isinstance(all_runs_canceled, bool):
            check.failed(
                "Expected cancel_backfill_runs_and_cancellation_complete to return a boolean"
            )

        if all_runs_canceled:
            instance.update_backfill(
                backfill.with_status(BulkActionStatus.CANCELED).with_end_timestamp(
                    get_current_timestamp()
                )
            )
        return

    has_more = True
    while has_more:
        if backfill.status != BulkActionStatus.REQUESTED:
            break

        chunk, checkpoint, has_more = _get_partitions_chunk(
            instance,
            logger,
            backfill,
            CHECKPOINT_COUNT,
            partition_set,
        )
        check_for_debug_crash(debug_crash_flags, "BEFORE_SUBMIT")

        if chunk:
            for _run_id in submit_backfill_runs(
                instance,
                lambda: workspace_process_context.create_request_context(),
                backfill,
                chunk,
            ):
                yield None
                # before submitting, refetch the backfill job to check for status changes
                backfill = cast(PartitionBackfill, instance.get_backfill(backfill.backfill_id))
                if backfill.status != BulkActionStatus.REQUESTED:
                    return

        check_for_debug_crash(debug_crash_flags, "AFTER_SUBMIT")

        if has_more:
            # refetch, in case the backfill was updated in the meantime
            backfill = cast(PartitionBackfill, instance.get_backfill(backfill.backfill_id))
            instance.update_backfill(backfill.with_partition_checkpoint(checkpoint))
            yield None
            time.sleep(CHECKPOINT_INTERVAL)
        else:
            unfinished_runs = instance.get_runs(
                RunsFilter(
                    tags=DagsterRun.tags_for_backfill_id(backfill.backfill_id),
                    statuses=NOT_FINISHED_STATUSES,
                ),
                limit=1,
            )
            if unfinished_runs:
                logger.info(
                    f"Backfill {backfill.backfill_id} has unfinished runs. Status will be updated when all runs are finished."
                )
                instance.update_backfill(backfill.with_partition_checkpoint(checkpoint))
                return
            partition_names = cast(Sequence[str], backfill.partition_names)
            logger.info(
                f"Backfill completed for {backfill.backfill_id} for"
                f" {len(partition_names)} partitions"
            )
            if (
                len(
                    instance.get_run_ids(
                        filters=RunsFilter(
                            tags=DagsterRun.tags_for_backfill_id(backfill.backfill_id),
                            statuses=[DagsterRunStatus.FAILURE, DagsterRunStatus.CANCELED],
                        )
                    )
                )
                > 0
            ):
                instance.update_backfill(
                    backfill.with_status(BulkActionStatus.COMPLETED_FAILED).with_end_timestamp(
                        get_current_timestamp()
                    )
                )
            else:
                instance.update_backfill(
                    backfill.with_status(BulkActionStatus.COMPLETED_SUCCESS).with_end_timestamp(
                        get_current_timestamp()
                    )
                )
            yield None


def _get_partition_set(
    workspace_process_context: IWorkspaceProcessContext, backfill_job: PartitionBackfill
) -> RemotePartitionSet:
    origin = cast(RemotePartitionSetOrigin, backfill_job.partition_set_origin)

    location_name = origin.repository_origin.code_location_origin.location_name

    workspace = workspace_process_context.create_request_context()
    code_location = workspace.get_code_location(location_name)

    repo_name = origin.repository_origin.repository_name
    if not code_location.has_repository(repo_name):
        raise DagsterBackfillFailedError(
            f"Could not find repository {repo_name} in location {code_location.name} to "
            f"run backfill {backfill_job.backfill_id}."
        )

    partition_set_name = origin.partition_set_name
    remote_repo = code_location.get_repository(repo_name)
    if not remote_repo.has_partition_set(partition_set_name):
        raise DagsterBackfillFailedError(
            f"Could not find partition set {partition_set_name} in repository {repo_name}. "
        )
    return remote_repo.get_partition_set(partition_set_name)


def _subdivide_partition_key_range(
    partitions_def: PartitionsDefinition,
    partition_key_range: PartitionKeyRange,
    instance: DagsterInstance,
    max_range_size: Optional[int],
) -> Sequence[PartitionKeyRange]:
    """Take a partition key range and subdivide it into smaller ranges of size max_range_size. This
    is done to satisfy backfill policies that limit the maximum number of partitions that can be
    materialized in a run.
    """
    if max_range_size is None:
        return [partition_key_range]
    else:
        keys = partitions_def.get_partition_keys_in_range(
            partition_key_range, dynamic_partitions_store=instance
        )
        chunks = [keys[i : i + max_range_size] for i in range(0, len(keys), max_range_size)]
        return [PartitionKeyRange(start=chunk[0], end=chunk[-1]) for chunk in chunks]


def _get_partitions_chunk(
    instance: DagsterInstance,
    logger: logging.Logger,
    backfill_job: PartitionBackfill,
    chunk_size: int,
    partition_set: RemotePartitionSet,
) -> tuple[Sequence[Union[str, PartitionKeyRange]], str, bool]:
    partition_names = cast(Sequence[str], backfill_job.partition_names)
    checkpoint = backfill_job.last_submitted_partition_name
    backfill_policy = partition_set.backfill_policy

    if (
        backfill_job.last_submitted_partition_name
        and backfill_job.last_submitted_partition_name in partition_names
    ):
        index = partition_names.index(backfill_job.last_submitted_partition_name)
        partition_names = partition_names[index + 1 :]

    # for idempotence, fetch all runs with the current backfill id
    backfill_runs = instance.get_runs(
        RunsFilter(tags=DagsterRun.tags_for_backfill_id(backfill_job.backfill_id))
    )
    # fetching the partitions def of a legacy dynamic partitioned op-job will raise an error
    # so guard against it by checking if the partitions def exists first
    partitions_def = (
        partition_set.get_partitions_definition()
        if partition_set.has_partitions_definition()
        else None
    )
    completed_partitions = []
    for run in backfill_runs:
        if (
            run.tags.get(ASSET_PARTITION_RANGE_START_TAG)
            and run.tags.get(ASSET_PARTITION_RANGE_END_TAG)
            and run.tags.get(PARTITION_NAME_TAG) is None
            and partitions_def is not None
        ):
            if partitions_def is None:
                # We should not hit this case, since all PartitionsDefinitions that can be put on
                # assets are fetchable via the ExternalPartitionSet. However, we do this check so that
                # we only fetch the partitions def once before the loop
                raise DagsterInvariantViolationError(
                    f"Cannot access PartitionsDefinition for backfill {backfill_job.backfill_id}. "
                )
            completed_partitions.extend(
                partitions_def.get_partition_keys_in_range(
                    PartitionKeyRange(
                        start=run.tags[ASSET_PARTITION_RANGE_START_TAG],
                        end=run.tags[ASSET_PARTITION_RANGE_END_TAG],
                    ),
                    instance,
                )
            )
        elif run.tags.get(PARTITION_NAME_TAG):
            completed_partitions.append(run.tags[PARTITION_NAME_TAG])

    initial_checkpoint = (
        partition_names.index(checkpoint) + 1 if checkpoint and checkpoint in partition_names else 0
    )
    partition_names = partition_names[initial_checkpoint:]
    if len(partition_names) == 0:
        # no more partitions to submit, return early
        return [], checkpoint or "", False

    if backfill_policy and backfill_policy.max_partitions_per_run != 1:
        to_submit = [
            partition_name
            for partition_name in partition_names
            if partition_name not in completed_partitions
        ]
        partitions_def = partition_set.get_partitions_definition()
        partitions_subset = partitions_def.subset_with_partition_keys(to_submit)
        partition_key_ranges = partitions_subset.get_partition_key_ranges(
            partitions_def, dynamic_partitions_store=instance
        )
        subdivided_ranges = [
            sr
            for r in partition_key_ranges
            for sr in _subdivide_partition_key_range(
                partitions_def, r, instance, backfill_policy.max_partitions_per_run
            )
        ]
        ranges_to_launch = subdivided_ranges[:chunk_size]
        has_more = chunk_size < len(subdivided_ranges)
        next_checkpoint = ranges_to_launch[-1].end if len(ranges_to_launch) > 0 else checkpoint
        to_submit = ranges_to_launch
    else:
        has_more = chunk_size < len(partition_names)
        partitions_chunk = partition_names[:chunk_size]
        next_checkpoint = partitions_chunk[-1]
        to_skip = set(partitions_chunk).intersection(completed_partitions)
        if to_skip:
            logger.info(
                f"Found {len(to_skip)} existing runs for backfill {backfill_job.backfill_id}, skipping"
            )
        to_submit = [
            partition_name
            for partition_name in partitions_chunk
            if partition_name not in completed_partitions
        ]

    return to_submit, next_checkpoint or "", has_more


def submit_backfill_runs(
    instance: DagsterInstance,
    create_workspace: Callable[[], BaseWorkspaceRequestContext],
    backfill_job: PartitionBackfill,
    partition_names_or_ranges: Optional[Sequence[Union[str, PartitionKeyRange]]] = None,
) -> Iterable[Optional[str]]:
    """Returns the run IDs of the submitted runs."""
    origin = cast(RemotePartitionSetOrigin, backfill_job.partition_set_origin)

    repository_origin = origin.repository_origin
    repo_name = repository_origin.repository_name
    location_name = repository_origin.code_location_origin.location_name

    if not partition_names_or_ranges:
        partition_names_or_ranges = cast(Sequence[str], backfill_job.partition_names)

    workspace = create_workspace()
    code_location = workspace.get_code_location(location_name)
    check.invariant(
        code_location.has_repository(repo_name),
        f"Could not find repository {repo_name} in location {code_location.name}",
    )
    remote_repo = code_location.get_repository(repo_name)
    partition_set_name = origin.partition_set_name
    partition_set = remote_repo.get_partition_set(partition_set_name)

    if backfill_job.asset_selection:
        # need to make another call to the user code location to properly subset
        # for an asset selection
        pipeline_selector = JobSubsetSelector(
            location_name=code_location.name,
            repository_name=repo_name,
            job_name=partition_set.job_name,
            op_selection=None,
            asset_selection=backfill_job.asset_selection,
        )
        remote_job = code_location.get_job(pipeline_selector)
    else:
        remote_job = remote_repo.get_full_job(partition_set.job_name)

    partition_data_target = check.is_list(
        [partition_names_or_ranges[0].start]
        if isinstance(partition_names_or_ranges[0], PartitionKeyRange)
        else partition_names_or_ranges,
        of_type=str,
    )
    partition_set_execution_data = code_location.get_partition_set_execution_params(
        remote_repo.handle,
        partition_set_name,
        partition_data_target,
        instance,
    )
    assert isinstance(partition_set_execution_data, PartitionSetExecutionParamSnap)

    # Partition-scoped run config is prohibited at the definitions level for a jobs that materialize
    # ranges, so we can assume that all partition data will have the same run config and tags as the
    # first partition.
    tags_by_key_or_range: Mapping[Union[str, PartitionKeyRange], Mapping[str, str]]
    run_config_by_key_or_range: Mapping[Union[str, PartitionKeyRange], Mapping[str, Any]]
    if isinstance(partition_names_or_ranges[0], PartitionKeyRange):
        run_config = partition_set_execution_data.partition_data[0].run_config
        tags = {
            k: v
            for k, v in partition_set_execution_data.partition_data[0].tags.items()
            if k != PARTITION_NAME_TAG
        }
        run_config_by_key_or_range = {r: run_config for r in partition_names_or_ranges}
        tags_by_key_or_range = {
            r: {
                **tags,
                ASSET_PARTITION_RANGE_START_TAG: r.start,
                ASSET_PARTITION_RANGE_END_TAG: r.end,
            }
            for r in check.is_list(partition_names_or_ranges, of_type=PartitionKeyRange)
        }
    else:
        run_config_by_key_or_range = {
            pd.name: pd.run_config for pd in partition_set_execution_data.partition_data
        }
        tags_by_key_or_range = {
            pd.name: pd.tags for pd in partition_set_execution_data.partition_data
        }

    for key_or_range in partition_names_or_ranges:
        # Refresh the code location in case the workspace has reloaded mid-backfill
        workspace = create_workspace()
        code_location = workspace.get_code_location(location_name)

        dagster_run = create_backfill_run(
            instance,
            code_location,
            remote_job,
            partition_set,
            backfill_job,
            key_or_range,
            run_tags=tags_by_key_or_range[key_or_range],
            run_config=run_config_by_key_or_range[key_or_range],
        )
        if dagster_run:
            # we skip runs in certain cases, e.g. we are running a `from_failure` backfill job
            # and the partition has had a successful run since the time the backfill was
            # scheduled
            instance.submit_run(dagster_run.run_id, workspace)
            yield dagster_run.run_id
        yield None


def create_backfill_run(
    instance: DagsterInstance,
    code_location: CodeLocation,
    remote_job: RemoteJob,
    remote_partition_set: RemotePartitionSet,
    backfill_job: PartitionBackfill,
    partition_key_or_range: Union[str, PartitionKeyRange],
    run_tags: Mapping[str, str],
    run_config: Mapping[str, Any],
) -> Optional[DagsterRun]:
    from dagster._daemon.daemon import get_telemetry_daemon_session_id

    log_action(
        instance,
        BACKFILL_RUN_CREATED,
        metadata={
            "DAEMON_SESSION_ID": get_telemetry_daemon_session_id(),
            "repo_hash": hash_name(code_location.name),
            "pipeline_name_hash": hash_name(remote_job.name),
        },
    )

    tags = merge_dicts(
        remote_job.tags,
        run_tags,
        DagsterRun.tags_for_backfill_id(backfill_job.backfill_id),
        backfill_job.tags,
    )

    resolved_op_selection = None
    op_selection = None
    if not backfill_job.from_failure and not backfill_job.reexecution_steps:
        step_keys_to_execute = None
        parent_run_id = None
        root_run_id = None
        known_state = None
        if remote_partition_set.op_selection:
            resolved_op_selection = frozenset(remote_partition_set.op_selection)
            op_selection = remote_partition_set.op_selection

    elif backfill_job.from_failure:
        last_run = _fetch_last_run(instance, remote_partition_set, partition_key_or_range)
        if not last_run or last_run.status != DagsterRunStatus.FAILURE:
            return None
        return instance.create_reexecuted_run(
            parent_run=last_run,
            code_location=code_location,
            remote_job=remote_job,
            strategy=ReexecutionStrategy.FROM_FAILURE,
            extra_tags=tags,
            run_config=run_config,
            use_parent_run_tags=False,  # don't inherit tags from the previous run
        )

    else:  # backfill_job.reexecution_steps
        last_run = _fetch_last_run(instance, remote_partition_set, partition_key_or_range)
        parent_run_id = last_run.run_id if last_run else None
        root_run_id = (last_run.root_run_id or last_run.run_id) if last_run else None
        if parent_run_id and root_run_id:
            tags = merge_dicts(
                tags, {PARENT_RUN_ID_TAG: parent_run_id, ROOT_RUN_ID_TAG: root_run_id}
            )
        step_keys_to_execute = backfill_job.reexecution_steps
        if last_run and last_run.status == DagsterRunStatus.SUCCESS:
            known_state = KnownExecutionState.build_for_reexecution(
                instance,
                last_run,
            ).update_for_step_selection(step_keys_to_execute)
        else:
            known_state = None

        if remote_partition_set.op_selection:
            resolved_op_selection = frozenset(remote_partition_set.op_selection)
            op_selection = remote_partition_set.op_selection

    remote_execution_plan = code_location.get_execution_plan(
        remote_job,
        run_config,
        step_keys_to_execute=step_keys_to_execute,
        known_state=known_state,
        instance=instance,
    )

    return instance.create_run(
        job_snapshot=remote_job.job_snapshot,
        execution_plan_snapshot=remote_execution_plan.execution_plan_snapshot,
        parent_job_snapshot=remote_job.parent_job_snapshot,
        job_name=remote_job.name,
        run_id=make_new_run_id(),
        resolved_op_selection=resolved_op_selection,
        run_config=run_config,
        step_keys_to_execute=step_keys_to_execute,
        tags=tags,
        root_run_id=root_run_id,
        parent_run_id=parent_run_id,
        status=DagsterRunStatus.NOT_STARTED,
        remote_job_origin=remote_job.get_remote_origin(),
        job_code_origin=remote_job.get_python_origin(),
        op_selection=op_selection,
        asset_selection=(
            frozenset(backfill_job.asset_selection) if backfill_job.asset_selection else None
        ),
        asset_check_selection=None,
        asset_graph=code_location.get_repository(
            remote_job.repository_handle.repository_name
        ).asset_graph,
    )


def _fetch_last_run(
    instance: DagsterInstance,
    remote_partition_set: RemotePartitionSet,
    partition_key_or_range: Union[str, PartitionKeyRange],
) -> Optional[DagsterRun]:
    check.inst_param(instance, "instance", DagsterInstance)
    check.inst_param(remote_partition_set, "remote_partition_set", RemotePartitionSet)
    check.str_param(partition_key_or_range, "partition_name")

    tags = (
        {
            PARTITION_NAME_TAG: partition_key_or_range,
        }
        if isinstance(partition_key_or_range, str)
        else {
            ASSET_PARTITION_RANGE_START_TAG: partition_key_or_range.start,
            ASSET_PARTITION_RANGE_END_TAG: partition_key_or_range.end,
        }
    )

    runs = instance.get_runs(
        RunsFilter(
            job_name=remote_partition_set.job_name,
            tags={PARTITION_SET_TAG: remote_partition_set.name, **tags},
        ),
        limit=1,
    )

    return runs[0] if runs else None



from dagster._model import DagsterModel
from dagster._serdes import whitelist_for_serdes


@whitelist_for_serdes()
class RunTelemetryData(DagsterModel):
    run_id: str
    datapoints: dict[str, float]



# -*- coding: utf-8 -*-
import warnings
import json

from tarfile import TarFile
from pkgutil import get_data
from io import BytesIO

# CHANGED IN VENDORED VERSION
from ..tz import tzfile as _tzfile

__all__ = ["get_zonefile_instance", "gettz", "gettz_db_metadata"]

ZONEFILENAME = "dateutil-zoneinfo.tar.gz"
METADATA_FN = 'METADATA'


class tzfile(_tzfile):
    def __reduce__(self):
        return (gettz, (self._filename,))


def getzoneinfofile_stream():
    try:
        return BytesIO(get_data(__name__, ZONEFILENAME))
    except IOError as e:  # TODO  switch to FileNotFoundError?
        warnings.warn("I/O error({0}): {1}".format(e.errno, e.strerror))
        return None


class ZoneInfoFile(object):
    def __init__(self, zonefile_stream=None):
        if zonefile_stream is not None:
            with TarFile.open(fileobj=zonefile_stream) as tf:
                self.zones = {zf.name: tzfile(tf.extractfile(zf), filename=zf.name)
                              for zf in tf.getmembers()
                              if zf.isfile() and zf.name != METADATA_FN}
                # deal with links: They'll point to their parent object. Less
                # waste of memory
                links = {zl.name: self.zones[zl.linkname]
                         for zl in tf.getmembers() if
                         zl.islnk() or zl.issym()}
                self.zones.update(links)
                try:
                    metadata_json = tf.extractfile(tf.getmember(METADATA_FN))
                    metadata_str = metadata_json.read().decode('UTF-8')
                    self.metadata = json.loads(metadata_str)
                except KeyError:
                    # no metadata in tar file
                    self.metadata = None
        else:
            self.zones = {}
            self.metadata = None

    def get(self, name, default=None):
        """
        Wrapper for :func:`ZoneInfoFile.zones.get`. This is a convenience method
        for retrieving zones from the zone dictionary.

        :param name:
            The name of the zone to retrieve. (Generally IANA zone names)

        :param default:
            The value to return in the event of a missing key.

        .. versionadded:: 2.6.0

        """
        return self.zones.get(name, default)


# The current API has gettz as a module function, although in fact it taps into
# a stateful class. So as a workaround for now, without changing the API, we
# will create a new "global" class instance the first time a user requests a
# timezone. Ugly, but adheres to the api.
#
# TODO: Remove after deprecation period.
_CLASS_ZONE_INSTANCE = []


def get_zonefile_instance(new_instance=False):
    """
    This is a convenience function which provides a :class:`ZoneInfoFile`
    instance using the data provided by the ``dateutil`` package. By default, it
    caches a single instance of the ZoneInfoFile object and returns that.

    :param new_instance:
        If ``True``, a new instance of :class:`ZoneInfoFile` is instantiated and
        used as the cached instance for the next call. Otherwise, new instances
        are created only as necessary.

    :return:
        Returns a :class:`ZoneInfoFile` object.

    .. versionadded:: 2.6
    """
    if new_instance:
        zif = None
    else:
        zif = getattr(get_zonefile_instance, '_cached_instance', None)

    if zif is None:
        zif = ZoneInfoFile(getzoneinfofile_stream())

        get_zonefile_instance._cached_instance = zif

    return zif


def gettz(name):
    """
    This retrieves a time zone from the local zoneinfo tarball that is packaged
    with dateutil.

    :param name:
        An IANA-style time zone name, as found in the zoneinfo file.

    :return:
        Returns a :class:`dateutil.tz.tzfile` time zone object.

    .. warning::
        It is generally inadvisable to use this function, and it is only
        provided for API compatibility with earlier versions. This is *not*
        equivalent to ``dateutil.tz.gettz()``, which selects an appropriate
        time zone based on the inputs, favoring system zoneinfo. This is ONLY
        for accessing the dateutil-specific zoneinfo (which may be out of
        date compared to the system zoneinfo).

    .. deprecated:: 2.6
        If you need to use a specific zoneinfofile over the system zoneinfo,
        instantiate a :class:`dateutil.zoneinfo.ZoneInfoFile` object and call
        :func:`dateutil.zoneinfo.ZoneInfoFile.get(name)` instead.

        Use :func:`get_zonefile_instance` to retrieve an instance of the
        dateutil-provided zoneinfo.
    """
    warnings.warn("zoneinfo.gettz() will be removed in future versions, "
                  "to use the dateutil-provided zoneinfo files, instantiate a "
                  "ZoneInfoFile object and use ZoneInfoFile.zones.get() "
                  "instead. See the documentation for details.",
                  DeprecationWarning)

    if len(_CLASS_ZONE_INSTANCE) == 0:
        _CLASS_ZONE_INSTANCE.append(ZoneInfoFile(getzoneinfofile_stream()))
    return _CLASS_ZONE_INSTANCE[0].zones.get(name)


def gettz_db_metadata():
    """ Get the zonefile metadata

    See `zonefile_metadata`_

    :returns:
        A dictionary with the database metadata

    .. deprecated:: 2.6
        See deprecation warning in :func:`zoneinfo.gettz`. To get metadata,
        query the attribute ``zoneinfo.ZoneInfoFile.metadata``.
    """
    warnings.warn("zoneinfo.gettz_db_metadata() will be removed in future "
                  "versions, to use the dateutil-provided zoneinfo files, "
                  "ZoneInfoFile object and query the 'metadata' attribute "
                  "instead. See the documentation for details.",
                  DeprecationWarning)

    if len(_CLASS_ZONE_INSTANCE) == 0:
        _CLASS_ZONE_INSTANCE.append(ZoneInfoFile(getzoneinfofile_stream()))
    return _CLASS_ZONE_INSTANCE[0].metadata



import os
from datetime import datetime
from enum import Enum
from typing import Optional

from dagster._record import record


def get_max_concurrency_limit_value() -> int:
    return int(os.getenv("DAGSTER_MAX_GLOBAL_OP_CONCURRENCY_LIMIT", "1000"))


class ConcurrencySlotStatus(Enum):
    BLOCKED = "BLOCKED"
    CLAIMED = "CLAIMED"


@record
class ConcurrencyClaimStatus:
    concurrency_key: str
    slot_status: ConcurrencySlotStatus
    priority: Optional[int] = None
    assigned_timestamp: Optional[datetime] = None
    enqueued_timestamp: Optional[datetime] = None
    sleep_interval: Optional[float] = None

    @property
    def is_claimed(self):
        return self.slot_status == ConcurrencySlotStatus.CLAIMED

    @property
    def is_assigned(self):
        return self.assigned_timestamp is not None

    def with_slot_status(self, slot_status: ConcurrencySlotStatus):
        return ConcurrencyClaimStatus(
            concurrency_key=self.concurrency_key,
            slot_status=slot_status,
            priority=self.priority,
            assigned_timestamp=self.assigned_timestamp,
            enqueued_timestamp=self.enqueued_timestamp,
            sleep_interval=self.sleep_interval,
        )

    def with_sleep_interval(self, interval: float):
        return ConcurrencyClaimStatus(
            concurrency_key=self.concurrency_key,
            slot_status=self.slot_status,
            priority=self.priority,
            assigned_timestamp=self.assigned_timestamp,
            enqueued_timestamp=self.enqueued_timestamp,
            sleep_interval=interval,
        )


@record
class PendingStepInfo:
    run_id: str
    step_key: str
    enqueued_timestamp: datetime
    assigned_timestamp: Optional[datetime]
    priority: Optional[int]


@record
class ClaimedSlotInfo:
    run_id: str
    step_key: str


@record
class ConcurrencyKeyInfo:
    concurrency_key: str
    slot_count: int
    claimed_slots: list[ClaimedSlotInfo]
    pending_steps: list[PendingStepInfo]
    # limit: None means no slots and no default limit.  `limit` will return the instance-configured default even if the slots have not yet been initialized.
    limit: Optional[int] = None
    using_default_limit: bool = False

    ###################################################
    # Fields that we need to keep around for backcompat
    ###################################################
    @property
    def active_slot_count(self) -> int:
        return len(self.claimed_slots)

    @property
    def active_run_ids(self) -> set[str]:
        return set([slot.run_id for slot in self.claimed_slots])

    @property
    def pending_step_count(self) -> int:
        # here pending steps are steps that are not assigned
        return len([step for step in self.pending_steps if step.assigned_timestamp is None])

    @property
    def pending_run_ids(self) -> set[str]:
        # here pending steps are steps that are not assigned
        return set([step.run_id for step in self.pending_steps if step.assigned_timestamp is None])

    @property
    def assigned_step_count(self) -> int:
        return len([step for step in self.pending_steps if step.assigned_timestamp is not None])

    @property
    def assigned_run_ids(self) -> set[str]:
        return set(
            [step.run_id for step in self.pending_steps if step.assigned_timestamp is not None]
        )



from dagster._core.definitions.declarative_automation.operators.any_downstream_conditions_operator import (
    AnyDownstreamConditionsCondition as AnyDownstreamConditionsCondition,
)
from dagster._core.definitions.declarative_automation.operators.boolean_operators import (
    AndAutomationCondition as AndAutomationCondition,
    NotAutomationCondition as NotAutomationCondition,
    OrAutomationCondition as OrAutomationCondition,
)
from dagster._core.definitions.declarative_automation.operators.check_operators import (
    AllChecksCondition as AllChecksCondition,
    AnyChecksCondition as AnyChecksCondition,
)
from dagster._core.definitions.declarative_automation.operators.dep_operators import (
    AllDepsCondition as AllDepsCondition,
    AnyDepsCondition as AnyDepsCondition,
    EntityMatchesCondition as EntityMatchesCondition,
)
from dagster._core.definitions.declarative_automation.operators.newly_true_operator import (
    NewlyTrueCondition as NewlyTrueCondition,
)
from dagster._core.definitions.declarative_automation.operators.since_operator import (
    SinceCondition as SinceCondition,
)



import base64
from collections.abc import Sequence
from typing import NamedTuple

from dagster._seven import json


class LogLineCursor(NamedTuple):
    """Representation of a log line cursor, to keep track of the place in the logs.
    The captured logs are stored in multiple files in the same direcotry. The cursor keeps
    track of the file name and the number of lines read so far.

    line=-1 means that the entire file has been read and the next file should be read. This covers the
    case when and entire file has been read, but the next file does not exist in storage yet.
    line=0 means no lines from the file have been read.
    line=n means lines 0 through n-1 have been read from the file.

    has_more_now indicates if there are more log lines that can be read immediately. If the process writing
    logs is still running, but has not writen a log file, has_more_now will be False once all currently readable
    log files have been read. It does not mean that no new logs will be written in the future.
    """

    log_key: Sequence[str]
    line: int  # maybe rename line_offset?
    has_more_now: bool

    def __str__(self) -> str:
        return self.to_string()

    def to_string(self) -> str:
        raw = json.dumps(
            {"log_key": self.log_key, "line": self.line, "has_more_now": self.has_more_now}
        )
        return base64.b64encode(bytes(raw, encoding="utf-8")).decode("utf-8")

    @staticmethod
    def parse(cursor_str: str) -> "LogLineCursor":
        raw = json.loads(base64.b64decode(cursor_str).decode("utf-8"))
        return LogLineCursor(raw["log_key"], raw["line"], raw["has_more_now"])



import typing

Any = typing.Any
Bool = bool
Float = float
Int = int
Nothing: type[None] = type(None)
String = str


class BuiltinEnum:
    ANY = Any
    BOOL = Bool
    FLOAT = Float
    INT = Int
    NOTHING = Nothing
    STRING = String

    @classmethod
    def contains(cls, value: object) -> bool:
        for ttype in [cls.ANY, cls.BOOL, cls.FLOAT, cls.INT, cls.STRING, cls.NOTHING]:
            if value == ttype:
                return True

        return False



import pickle

import dagster._check as check
import dagster._seven as seven
from dagster._config import (
    ConfigAnyInstance,
    ConfigBoolInstance,
    ConfigFloatInstance,
    ConfigIntInstance,
    ConfigStringInstance,
    ConfigType,
    ConfigTypeKind,
    Field,
    ScalarUnion,
    Selector,
)
from dagster._core.types.config_schema import dagster_type_loader
from dagster._utils.warnings import disable_dagster_warnings


def define_typed_input_schema_dict(value_config_type):
    check.inst_param(value_config_type, "value_config_type", ConfigType)
    return Selector(
        {
            "value": Field(value_config_type),
            "json": define_path_dict_field(),
            "pickle": define_path_dict_field(),
        },
    )


def load_type_input_schema_dict(value):
    file_type, file_options = next(iter(value.items()))
    if file_type == "value":
        return file_options
    elif file_type == "json":
        with open(file_options["path"], "r", encoding="utf8") as ff:
            value_dict = seven.json.load(ff)
            return value_dict["value"]
    elif file_type == "pickle":
        with open(file_options["path"], "rb") as ff:
            return pickle.load(ff)
    else:
        check.failed(f"Unsupported key {file_type}")


def define_any_input_schema():
    @dagster_type_loader(define_typed_input_schema_dict(ConfigAnyInstance))
    def _any_input_schema(_, config_value):
        return load_type_input_schema_dict(config_value)

    return _any_input_schema


def define_builtin_scalar_input_schema(scalar_name, config_scalar_type):
    check.str_param(scalar_name, "scalar_name")
    check.inst_param(config_scalar_type, "config_scalar_type", ConfigType)
    check.param_invariant(
        config_scalar_type.kind == ConfigTypeKind.SCALAR, "config_scalar_type"
    )
    # TODO: https://github.com/dagster-io/dagster/issues/3084
    with disable_dagster_warnings():

        @dagster_type_loader(
            ScalarUnion(
                scalar_type=config_scalar_type,
                non_scalar_schema=define_typed_input_schema_dict(config_scalar_type),
            ),
        )
        def _builtin_input_schema(_context, value):
            return (
                load_type_input_schema_dict(value) if isinstance(value, dict) else value
            )

    return _builtin_input_schema


def define_path_dict_field():
    return {"path": Field(ConfigStringInstance)}


class BuiltinSchemas:
    ANY_INPUT = define_any_input_schema()

    BOOL_INPUT = define_builtin_scalar_input_schema("Bool", ConfigBoolInstance)

    FLOAT_INPUT = define_builtin_scalar_input_schema("Float", ConfigFloatInstance)

    INT_INPUT = define_builtin_scalar_input_schema("Int", ConfigIntInstance)

    STRING_INPUT = define_builtin_scalar_input_schema("String", ConfigStringInstance)



from typing import Optional


def is_basic_daily(cron_schedule: str) -> bool:
    return cron_schedule == "0 0 * * *"


def is_basic_hourly(cron_schedule: str) -> bool:
    return cron_schedule == "0 * * * *"


def get_fixed_minute_interval(cron_schedule: str) -> Optional[int]:
    """Given a cronstring, returns whether or not it is safe to
    assume there is a fixed number of minutes between every tick. For
    many cronstrings this is not the case due to Daylight Savings Time,
    but for basic hourly cron schedules and cron schedules like */15 it
    is safe to assume that there are a fixed number of minutes between each
    tick.
    """
    if is_basic_hourly(cron_schedule):
        return 60

    cron_parts = cron_schedule.split()
    is_wildcard = [part == "*" for part in cron_parts]

    # To match this criteria, every other field besides the first must end in *
    # since it must be an every-n-minutes cronstring like */15
    if not is_wildcard[1:]:
        return None

    if not cron_parts[0].startswith("*/"):
        return None

    try:
        # interval makes up the characters after the "*/"
        interval = int(cron_parts[0][2:])
    except ValueError:
        return None

    # cronstrings like */7 do not have a fixed interval because they jump
    # from :54 to :07, but divisors of 60 do
    if interval > 0 and interval < 60 and 60 % interval == 0:
        return interval

    return None



from collections.abc import Mapping, Sequence
from typing import Any, Optional, Union

import dagster._check as check
from dagster._annotations import hidden_param, only_allow_hidden_params_in_kwargs, public
from dagster._core.definitions.asset_dep import AssetDep
from dagster._core.definitions.asset_spec import AssetSpec
from dagster._core.definitions.auto_materialize_policy import AutoMaterializePolicy
from dagster._core.definitions.backfill_policy import BackfillPolicy
from dagster._core.definitions.declarative_automation.automation_condition import (
    AutomationCondition,
)
from dagster._core.definitions.events import (
    AssetKey,
    CoercibleToAssetKey,
    CoercibleToAssetKeyPrefix,
)
from dagster._core.definitions.freshness_policy import FreshnessPolicy
from dagster._core.definitions.input import NoValueSentinel
from dagster._core.definitions.output import Out
from dagster._core.definitions.partition import PartitionsDefinition
from dagster._core.definitions.utils import resolve_automation_condition
from dagster._core.errors import DagsterInvalidDefinitionError
from dagster._core.types.dagster_type import DagsterType
from dagster._utils.tags import normalize_tags
from dagster._utils.warnings import disable_dagster_warnings

# Unfortunate, since AssetOut stores nearly all of the properties of
# an AssetSpec except for the key, we use a sentinel value to represent
# an unspecified key
EMPTY_ASSET_KEY_SENTINEL = AssetKey([])


@hidden_param(
    param="freshness_policy",
    breaking_version="1.10.0",
    additional_warn_text="use freshness checks instead",
)
@hidden_param(
    param="auto_materialize_policy",
    breaking_version="1.10.0",
    additional_warn_text="use `automation_condition` instead",
)
class AssetOut:
    """Defines one of the assets produced by a :py:func:`@multi_asset <multi_asset>`.

    Args:
        key_prefix (Optional[Union[str, Sequence[str]]]): If provided, the asset's key is the
            concatenation of the key_prefix and the asset's name. When using ``@multi_asset``, the
            asset name defaults to the key of the "outs" dictionary Only one of the "key_prefix" and
            "key" arguments should be provided.
        key (Optional[Union[str, Sequence[str], AssetKey]]): The asset's key. Only one of the
            "key_prefix" and "key" arguments should be provided.
        dagster_type (Optional[Union[Type, DagsterType]]]):
            The type of this output. Should only be set if the correct type can not
            be inferred directly from the type signature of the decorated function.
        description (Optional[str]): Human-readable description of the output.
        is_required (bool): Whether the presence of this field is required. (default: True)
        io_manager_key (Optional[str]): The resource key of the IO manager used for this output.
            (default: "io_manager").
        metadata (Optional[Dict[str, Any]]): A dict of the metadata for the output.
            For example, users can provide a file path if the data object will be stored in a
            filesystem, or provide information of a database table when it is going to load the data
            into the table.
        group_name (Optional[str]): A string name used to organize multiple assets into groups. If
            not provided, the name "default" is used.
        code_version (Optional[str]): The version of the code that generates this asset.
        freshness_policy (Optional[FreshnessPolicy]): (Deprecated) A policy which indicates how up
            to date this asset is intended to be.
        automation_condition (Optional[AutomationCondition]): AutomationCondition to apply to the
            specified asset.
        backfill_policy (Optional[BackfillPolicy]): BackfillPolicy to apply to the specified asset.
        owners (Optional[Sequence[str]]): A list of strings representing owners of the asset. Each
            string can be a user's email address, or a team name prefixed with `team:`,
            e.g. `team:finops`.
        tags (Optional[Mapping[str, str]]): Tags for filtering and organizing. These tags are not
            attached to runs of the asset.
    """

    _spec: AssetSpec
    key_prefix: Optional[Sequence[str]]
    dagster_type: Union[type, DagsterType]
    is_required: bool
    io_manager_key: Optional[str]
    backfill_policy: Optional[BackfillPolicy]

    def __init__(
        self,
        key_prefix: Optional[CoercibleToAssetKeyPrefix] = None,
        key: Optional[CoercibleToAssetKey] = None,
        dagster_type: Union[type, DagsterType] = NoValueSentinel,
        description: Optional[str] = None,
        is_required: bool = True,
        io_manager_key: Optional[str] = None,
        metadata: Optional[Mapping[str, Any]] = None,
        group_name: Optional[str] = None,
        code_version: Optional[str] = None,
        automation_condition: Optional[AutomationCondition] = None,
        backfill_policy: Optional[BackfillPolicy] = None,
        owners: Optional[Sequence[str]] = None,
        tags: Optional[Mapping[str, str]] = None,
        **kwargs,
    ):
        # Accept a hidden "spec" argument to allow for the AssetOut to be constructed from an AssetSpec
        # directly. This is used in the AssetOut.from_spec method.
        spec = kwargs.get("spec")
        if spec:
            del kwargs["spec"]

        only_allow_hidden_params_in_kwargs(AssetOut, kwargs)
        if isinstance(key_prefix, str):
            key_prefix = [key_prefix]

        check.invariant(
            not (key_prefix and key), "Only one of key_prefix and key should be provided"
        )

        auto_materialize_policy = kwargs.get("auto_materialize_policy")
        freshness_policy = kwargs.get("freshness_policy")
        has_any_spec_args = any(
            [
                key,
                description,
                metadata,
                group_name,
                code_version,
                automation_condition,
                auto_materialize_policy,
                freshness_policy,
                owners,
                tags,
            ]
        )
        check.invariant(
            not has_any_spec_args or not spec,
            "Cannot provide both spec and spec-related arguments (key, description, metadata, etc.)",
        )

        with disable_dagster_warnings():
            # This is a bit of a hack, since technically AssetOut does not hold all of the
            # properties of an AssetSpec - chiefly, it is missing a key. Still, this reduces
            # the amount of code duplication storing each of the properties in both AssetOut
            # and AssetSpec, and allows us to implement the from_spec method below.
            self._spec = spec or AssetSpec(
                key=AssetKey.from_coercible(key) if key is not None else EMPTY_ASSET_KEY_SENTINEL,
                description=check.opt_str_param(description, "description"),
                metadata=check.opt_mapping_param(metadata, "metadata", key_type=str),
                group_name=check.opt_str_param(group_name, "group_name"),
                code_version=check.opt_str_param(code_version, "code_version"),
                automation_condition=check.opt_inst_param(
                    resolve_automation_condition(automation_condition, auto_materialize_policy),
                    "automation_condition",
                    AutomationCondition,
                ),
                freshness_policy=check.opt_inst_param(
                    freshness_policy, "freshness_policy", FreshnessPolicy
                ),
                owners=check.opt_sequence_param(owners, "owners", of_type=str),
                tags=normalize_tags(tags or {}, strict=True),
            )
        self.key_prefix = key_prefix
        self.dagster_type = dagster_type
        self.is_required = is_required
        self.io_manager_key = io_manager_key
        self.backfill_policy = backfill_policy

    @property
    def key(self) -> Optional[AssetKey]:
        return self._spec.key if self._spec.key != EMPTY_ASSET_KEY_SENTINEL else None

    @property
    def metadata(self) -> Optional[Mapping[str, Any]]:
        return self._spec.metadata

    @property
    def description(self) -> Optional[str]:
        return self._spec.description

    @property
    def group_name(self) -> Optional[str]:
        return self._spec.group_name

    @property
    def code_version(self) -> Optional[str]:
        return self._spec.code_version

    @property
    def freshness_policy(self) -> Optional[FreshnessPolicy]:
        return self._spec.freshness_policy

    @property
    def automation_condition(self) -> Optional[AutomationCondition]:
        return self._spec.automation_condition

    @property
    def owners(self) -> Optional[Sequence[str]]:
        return self._spec.owners

    @property
    def tags(self) -> Optional[Mapping[str, str]]:
        return self._spec.tags

    def to_out(self) -> Out:
        return Out(
            dagster_type=self.dagster_type,
            description=self.description,
            metadata=self.metadata,
            is_required=self.is_required,
            io_manager_key=self.io_manager_key,
            code_version=self.code_version,
        )

    def to_spec(
        self,
        key: AssetKey,
        deps: Sequence[AssetDep],
        additional_tags: Mapping[str, str] = {},
        partitions_def: Optional[PartitionsDefinition] = ...,
    ) -> AssetSpec:
        return self._spec.replace_attributes(
            key=key,
            tags={**additional_tags, **self.tags} if self.tags else additional_tags,
            deps=[*self._spec.deps, *deps],
            partitions_def=partitions_def if partitions_def is not None else ...,
        )

    @public
    @staticmethod
    def from_spec(
        spec: AssetSpec,
        dagster_type: Union[type, DagsterType] = NoValueSentinel,
        is_required: bool = True,
        io_manager_key: Optional[str] = None,
        backfill_policy: Optional[BackfillPolicy] = None,
    ) -> "AssetOut":
        """Builds an AssetOut from the passed spec.

        Args:
            spec (AssetSpec): The spec to build the AssetOut from.
            dagster_type (Optional[Union[Type, DagsterType]]): The type of this output. Should only
                be set if the correct type can not be inferred directly from the type signature of
                the decorated function.
            is_required (bool): Whether the presence of this field is required. (default: True)
            io_manager_key (Optional[str]): The resource key of the IO manager used for this output.
                (default: "io_manager").
            backfill_policy (Optional[BackfillPolicy]): BackfillPolicy to apply to the specified
                asset.

        Returns:
            AssetOut: The AssetOut built from the spec.
        """
        if spec.deps:
            raise DagsterInvalidDefinitionError(
                "Currently, cannot build AssetOut from spec with deps"
            )
        return AssetOut(
            spec=spec,
            dagster_type=dagster_type,
            is_required=is_required,
            io_manager_key=io_manager_key,
            backfill_policy=backfill_policy,
        )

    @property
    def auto_materialize_policy(self) -> Optional[AutoMaterializePolicy]:
        return (
            self.automation_condition.as_auto_materialize_policy()
            if self.automation_condition
            else None
        )



import datetime
from collections.abc import Iterable, Sequence
from typing import Any, Optional, Union, cast

from dagster import _check as check
from dagster._annotations import beta
from dagster._core.definitions.asset_check_factories.utils import (
    DEADLINE_CRON_PARAM_KEY,
    DEFAULT_FRESHNESS_SEVERITY,
    DEFAULT_FRESHNESS_TIMEZONE,
    FRESH_UNTIL_METADATA_KEY,
    FRESHNESS_PARAMS_METADATA_KEY,
    LAST_UPDATED_TIMESTAMP_METADATA_KEY,
    LATEST_CRON_TICK_METADATA_KEY,
    LOWER_BOUND_DELTA_PARAM_KEY,
    LOWER_BOUND_TIMESTAMP_METADATA_KEY,
    TIMEZONE_PARAM_KEY,
    assets_to_keys,
    ensure_no_duplicate_assets,
    freshness_multi_asset_check,
    retrieve_last_update_record,
    retrieve_timestamp_from_record,
    seconds_in_words,
)
from dagster._core.definitions.asset_check_result import AssetCheckResult
from dagster._core.definitions.asset_check_spec import AssetCheckSeverity
from dagster._core.definitions.asset_checks import AssetChecksDefinition
from dagster._core.definitions.assets import AssetsDefinition, SourceAsset
from dagster._core.definitions.events import AssetKey, CoercibleToAssetKey
from dagster._core.definitions.metadata import (
    JsonMetadataValue,
    MetadataValue,
    TimestampMetadataValue,
)
from dagster._core.execution.context.compute import AssetCheckExecutionContext
from dagster._time import get_current_timestamp, get_timezone
from dagster._utils.schedules import (
    get_latest_completed_cron_tick,
    get_next_cron_tick,
    is_valid_cron_string,
)


@beta
def build_last_update_freshness_checks(
    *,
    assets: Sequence[Union[CoercibleToAssetKey, AssetsDefinition, SourceAsset]],
    lower_bound_delta: datetime.timedelta,
    deadline_cron: Optional[str] = None,
    timezone: str = DEFAULT_FRESHNESS_TIMEZONE,
    severity: AssetCheckSeverity = DEFAULT_FRESHNESS_SEVERITY,
) -> Sequence[AssetChecksDefinition]:
    r"""Constructs an `AssetChecksDefinition` that checks the freshness of the provided assets.

    This check passes if the asset is found to be "fresh", and fails if the asset is found to be
    "overdue". An asset is considered fresh if a record (i.e. a materialization or observation)
    exists with a timestamp greater than the "lower bound" derived from the parameters of this
    function.

    `deadline_cron` is a cron schedule that defines the deadline for when we should expect the asset
    to arrive by; if not provided, we consider the deadline to be the execution time of the check.
    `lower_bound_delta` is a timedelta that defines the lower bound for when a record could have
    arrived by. If the most recent recent record's timestamp is earlier than
    `deadline-lower_bound_delta`, the asset is considered overdue.

    Let's use two examples, one with a deadline_cron set and one without.
    Let's say I have an asset which runs on a schedule every day at 8:00 AM UTC, and usually takes
    around 45 minutes to complete. To account for operational delays, I would expect the asset to be
    done materializing every day by 9:00 AM UTC. I would set the `deadline_cron` to "0 9 \* \* \*", and
    the `lower_bound_delta` to "45 minutes". This would mean that starting at 9:00 AM, this check
    will expect a materialization record to have been created no earlier than 8:15 AM. Note that if
    the check runs at 8:59 AM, the deadline has not yet passed, and we'll instead be checking for
    the most recently passed deadline, which is yesterday.
    Let's say I have an observable source asset on a data source which I expect should never be more
    than 3 hours out of date. In this case, there's no fixed schedule for when the data should be
    updated, so I would not provide a `deadline_cron`. Instead, I would set the `lower_bound_delta`
    parameter to "3 hours". This would mean that the check will expect the most recent observation
    record to indicate data no older than 3 hours, relative to the current time, regardless of when it runs.

    The check result will contain the following metadata:
    "dagster/freshness_params": A dictionary containing the parameters used to construct the
    check
    "dagster/last_updated_time": The time of the most recent update to the asset
    "dagster/overdue_seconds": (Only present if asset is overdue) The number of seconds that the
    asset is overdue by.
    "dagster/overdue_deadline_timestamp": The timestamp that we are expecting the asset to have
    arrived by. In the case of a provided deadline_cron, this is the timestamp of the most recent
    tick of the cron schedule. In the case of no deadline_cron, this is the current time.

    Examples:
        .. code-block:: python

            # Example 1: Assets that are expected to be updated every day within 45 minutes of
            # 9:00 AM UTC
            from dagster import build_last_update_freshness_checks, AssetKey
            from .somewhere import my_daily_scheduled_assets_def

            checks_def = build_last_update_freshness_checks(
                [my_daily_scheduled_assets_def, AssetKey("my_other_daily_asset_key")],
                lower_bound_delta=datetime.timedelta(minutes=45),
                deadline_cron="0 9 * * *",
            )

            # Example 2: Assets that are expected to be updated within 3 hours of the current time
            from dagster import build_last_update_freshness_checks, AssetKey
            from .somewhere import my_observable_source_asset

            checks_def = build_last_update_freshness_checks(
                [my_observable_source_asset, AssetKey("my_other_observable_asset_key")],
                lower_bound_delta=datetime.timedelta(hours=3),
            )


    Args:
        assets (Sequence[Union[CoercibleToAssetKey, AssetsDefinition, SourceAsset]): The assets to
            construct checks for. All checks are incorporated into the same `AssetChecksDefinition`,
            which can be subsetted to run checks for specific assets.
        lower_bound_delta (datetime.timedelta): The check will pass if the asset was updated within
            lower_bound_delta of the current_time (no cron) or the most recent tick of the cron
            (cron provided).
        deadline_cron (Optional[str]): Defines the deadline for when we should start checking
            that the asset arrived. If not provided, the deadline is the execution time of the check.
        timezone (Optional[str]): The timezone to use when calculating freshness and deadline. If
            not provided, defaults to "UTC".

    Returns:
        Sequence[AssetChecksDefinition]: `AssetChecksDefinition` objects which execute freshness checks
            for the provided assets.
    """
    check.inst_param(lower_bound_delta, "lower_bound_delta", datetime.timedelta)
    check.str_param(timezone, "timezone")
    check.opt_str_param(deadline_cron, "deadline_cron")
    check.invariant(
        deadline_cron is None or is_valid_cron_string(deadline_cron), "Invalid cron string."
    )
    check.inst_param(severity, "severity", AssetCheckSeverity)
    check.sequence_param(assets, "assets")
    ensure_no_duplicate_assets(assets)
    return [
        _build_freshness_multi_check(
            asset_keys=assets_to_keys(assets),
            deadline_cron=deadline_cron,
            timezone=timezone,
            severity=severity,
            lower_bound_delta=lower_bound_delta,
        )
    ]


def _build_freshness_multi_check(
    asset_keys: Sequence[AssetKey],
    deadline_cron: Optional[str],
    timezone: str,
    severity: AssetCheckSeverity,
    lower_bound_delta: datetime.timedelta,
) -> AssetChecksDefinition:
    params_metadata: dict[str, Any] = {
        TIMEZONE_PARAM_KEY: timezone,
        LOWER_BOUND_DELTA_PARAM_KEY: lower_bound_delta.total_seconds(),
    }
    if deadline_cron:
        params_metadata[DEADLINE_CRON_PARAM_KEY] = deadline_cron

    @freshness_multi_asset_check(
        params_metadata=JsonMetadataValue(params_metadata), asset_keys=asset_keys
    )
    def the_check(context: AssetCheckExecutionContext) -> Iterable[AssetCheckResult]:
        for check_key in context.selected_asset_check_keys:
            asset_key = check_key.asset_key
            current_timestamp = get_current_timestamp()

            current_time_in_freshness_tz = datetime.datetime.fromtimestamp(
                current_timestamp, tz=get_timezone(timezone)
            )
            latest_completed_cron_tick = (
                get_latest_completed_cron_tick(
                    deadline_cron, current_time_in_freshness_tz, timezone
                )
                if deadline_cron
                else None
            )
            deadline = check.inst_param(
                latest_completed_cron_tick or current_time_in_freshness_tz,
                "deadline",
                datetime.datetime,
            )

            last_update_time_lower_bound = cast(datetime.datetime, deadline - lower_bound_delta)

            latest_record = retrieve_last_update_record(
                instance=context.instance, asset_key=asset_key, partition_key=None
            )
            update_timestamp = (
                retrieve_timestamp_from_record(latest_record) if latest_record else None
            )
            passed = (
                update_timestamp is not None
                and update_timestamp >= last_update_time_lower_bound.timestamp()
            )

            metadata: dict[str, MetadataValue] = {
                FRESHNESS_PARAMS_METADATA_KEY: JsonMetadataValue(params_metadata),
                LOWER_BOUND_TIMESTAMP_METADATA_KEY: TimestampMetadataValue(
                    last_update_time_lower_bound.timestamp()
                ),
            }
            if latest_completed_cron_tick:
                metadata[LATEST_CRON_TICK_METADATA_KEY] = TimestampMetadataValue(
                    latest_completed_cron_tick.timestamp()
                )
            if passed:
                # If the asset is fresh, we can determine when it has the possibility of becoming stale again.
                # In the case of a deadline cron, this is the next cron tick after the current time.
                # In the case of just a lower_bound_delta, this is the last update time plus the
                # lower_bound_delta.
                fresh_until = (
                    get_next_cron_tick(
                        deadline_cron, current_time_in_freshness_tz, timezone
                    ).timestamp()
                    if deadline_cron
                    else check.not_none(update_timestamp) + lower_bound_delta.total_seconds()
                )
                metadata[FRESH_UNTIL_METADATA_KEY] = TimestampMetadataValue(fresh_until)
            if update_timestamp:
                metadata[LAST_UPDATED_TIMESTAMP_METADATA_KEY] = TimestampMetadataValue(
                    update_timestamp
                )

            yield AssetCheckResult(
                passed=passed,
                description=construct_description(
                    passed,
                    last_update_time_lower_bound=last_update_time_lower_bound.timestamp(),
                    current_timestamp=current_timestamp,
                    update_timestamp=update_timestamp,
                ),
                severity=severity,
                asset_key=asset_key,
                metadata=metadata,
            )

    return the_check


def construct_description(
    passed: bool,
    last_update_time_lower_bound: Optional[float],
    current_timestamp: float,
    update_timestamp: Optional[float],
) -> str:
    check.invariant(
        (passed and update_timestamp is not None) or not passed,
        "Should not be possible for check to pass without an update to the asset.",
    )

    allowed_delta = (
        seconds_in_words(current_timestamp - last_update_time_lower_bound)
        if last_update_time_lower_bound
        else None
    )
    actual_delta = (
        seconds_in_words(current_timestamp - update_timestamp) if update_timestamp else None
    )

    if passed:
        return f"Asset is currently fresh. The last update was {actual_delta} ago, within the allowed time range of {allowed_delta}."
    elif update_timestamp is None:
        return "Asset has never been observed/materialized."
    else:
        return (
            f"Asset is overdue for an update. The last update was {actual_delta} ago, "
            f"which is past the allowed distance of {allowed_delta} ago."
        )



import base64
from collections.abc import Mapping, Sequence
from datetime import datetime
from enum import Enum
from typing import Callable, Literal, NamedTuple, Optional, Union

from typing_extensions import TypeAlias

import dagster._check as check
from dagster._annotations import PublicAttr
from dagster._core.definitions.events import AssetKey, AssetMaterialization, AssetObservation
from dagster._core.events import EVENT_TYPE_TO_PIPELINE_RUN_STATUS, DagsterEventType
from dagster._core.events.log import EventLogEntry
from dagster._serdes import whitelist_for_serdes
from dagster._seven import json

EventHandlerFn: TypeAlias = Callable[[EventLogEntry, str], None]


class EventLogCursorType(Enum):
    OFFSET = "OFFSET"
    STORAGE_ID = "STORAGE_ID"


class EventLogCursor(NamedTuple):
    """Representation of an event record cursor, keeping track of the log query state."""

    cursor_type: EventLogCursorType
    value: int

    def is_offset_cursor(self) -> bool:
        return self.cursor_type == EventLogCursorType.OFFSET

    def is_id_cursor(self) -> bool:
        return self.cursor_type == EventLogCursorType.STORAGE_ID

    def offset(self) -> int:
        check.invariant(self.cursor_type == EventLogCursorType.OFFSET)
        return max(0, int(self.value))

    def storage_id(self) -> int:
        check.invariant(self.cursor_type == EventLogCursorType.STORAGE_ID)
        return int(self.value)

    def __str__(self) -> str:
        return self.to_string()

    def to_string(self) -> str:
        raw = json.dumps({"type": self.cursor_type.value, "value": self.value})
        return base64.b64encode(bytes(raw, encoding="utf-8")).decode("utf-8")

    @staticmethod
    def parse(cursor_str: str) -> "EventLogCursor":
        raw = json.loads(base64.b64decode(cursor_str).decode("utf-8"))
        return EventLogCursor(EventLogCursorType(raw["type"]), raw["value"])

    @staticmethod
    def from_offset(offset: int) -> "EventLogCursor":
        return EventLogCursor(EventLogCursorType.OFFSET, offset)

    @staticmethod
    def from_storage_id(storage_id: int) -> "EventLogCursor":
        return EventLogCursor(EventLogCursorType.STORAGE_ID, storage_id)


class RunShardedEventsCursor(NamedTuple):
    """Pairs an id-based event log cursor with a timestamp-based run cursor, for improved
    performance on run-sharded event log storages (e.g. the default SqliteEventLogStorage). For
    run-sharded storages, the id field is ignored, since they may not be unique across shards.
    """

    id: int
    run_updated_after: datetime


RunStatusChangeEventType: TypeAlias = Literal[
    DagsterEventType.RUN_START,
    DagsterEventType.RUN_SUCCESS,
    DagsterEventType.RUN_FAILURE,
    DagsterEventType.RUN_ENQUEUED,
    DagsterEventType.RUN_STARTING,
    DagsterEventType.RUN_CANCELING,
    DagsterEventType.RUN_CANCELED,
]
AssetEventType: TypeAlias = Literal[
    DagsterEventType.ASSET_MATERIALIZATION,
    DagsterEventType.ASSET_OBSERVATION,
    DagsterEventType.ASSET_MATERIALIZATION_PLANNED,
]

EventCursor: TypeAlias = Union[int, RunShardedEventsCursor]


@whitelist_for_serdes
class EventLogRecord(NamedTuple):
    """Internal representation of an event record, as stored in a
    :py:class:`~dagster._core.storage.event_log.EventLogStorage`.

    Users should not instantiate this class directly.
    """

    storage_id: PublicAttr[int]
    event_log_entry: PublicAttr[EventLogEntry]

    @property
    def run_id(self) -> str:
        return self.event_log_entry.run_id

    @property
    def timestamp(self) -> float:
        return self.event_log_entry.timestamp

    @property
    def asset_key(self) -> Optional[AssetKey]:
        dagster_event = self.event_log_entry.dagster_event
        if dagster_event:
            return dagster_event.asset_key

        return None

    @property
    def partition_key(self) -> Optional[str]:
        dagster_event = self.event_log_entry.dagster_event
        if dagster_event:
            return dagster_event.partition

        return None

    @property
    def asset_materialization(self) -> Optional[AssetMaterialization]:
        return self.event_log_entry.asset_materialization

    @property
    def asset_observation(self) -> Optional[AssetObservation]:
        return self.event_log_entry.asset_observation

    @property
    def asset_event(self) -> Optional[Union[AssetMaterialization, AssetObservation]]:
        return self.asset_materialization or self.asset_observation

    @property
    def event_type(self) -> DagsterEventType:
        return check.not_none(
            self.event_log_entry.dagster_event,
            "Expected dagster_event property to be present if calling the event_type property",
        ).event_type


class EventRecordsResult(NamedTuple):
    """Return value for a query fetching event records from the instance.  Contains a list of event
    records, a cursor string, and a boolean indicating whether there are more records to fetch.
    """

    records: Sequence[EventLogRecord]
    cursor: str
    has_more: bool


@whitelist_for_serdes
class EventRecordsFilter(
    NamedTuple(
        "_EventRecordsFilter",
        [
            ("event_type", DagsterEventType),
            ("asset_key", Optional[AssetKey]),
            ("asset_partitions", Optional[Sequence[str]]),
            ("after_cursor", Optional[EventCursor]),
            ("before_cursor", Optional[EventCursor]),
            ("after_timestamp", Optional[float]),
            ("before_timestamp", Optional[float]),
            ("storage_ids", Optional[Sequence[int]]),
        ],
    )
):
    """Defines a set of filter fields for fetching a set of event log entries or event log records.

    Args:
        event_type (DagsterEventType): Filter argument for dagster event type
        asset_key (Optional[AssetKey]): Asset key for which to get asset materialization event
            entries / records.
        asset_partitions (Optional[List[str]]): Filter parameter such that only asset
            events with a partition value matching one of the provided values.  Only
            valid when the `asset_key` parameter is provided.
        after_cursor (Optional[EventCursor]): Filter parameter such that only
            records with storage_id greater than the provided value are returned. Using a
            run-sharded events cursor will result in a significant performance gain when run against
            a SqliteEventLogStorage implementation (which is run-sharded)
        before_cursor (Optional[EventCursor]): Filter parameter such that
            records with storage_id less than the provided value are returned. Using a run-sharded
            events cursor will result in a significant performance gain when run against
            a SqliteEventLogStorage implementation (which is run-sharded)
        after_timestamp (Optional[float]): Filter parameter such that only event records for
            events with timestamp greater than the provided value are returned.
        before_timestamp (Optional[float]): Filter parameter such that only event records for
            events with timestamp less than the provided value are returned.
    """

    def __new__(
        cls,
        event_type: DagsterEventType,
        asset_key: Optional[AssetKey] = None,
        asset_partitions: Optional[Sequence[str]] = None,
        after_cursor: Optional[EventCursor] = None,
        before_cursor: Optional[EventCursor] = None,
        after_timestamp: Optional[float] = None,
        before_timestamp: Optional[float] = None,
        storage_ids: Optional[Sequence[int]] = None,
    ):
        check.opt_sequence_param(asset_partitions, "asset_partitions", of_type=str)
        check.inst_param(event_type, "event_type", DagsterEventType)

        # type-ignores work around mypy type inference bug
        return super().__new__(
            cls,
            event_type=event_type,
            asset_key=check.opt_inst_param(asset_key, "asset_key", AssetKey),
            asset_partitions=asset_partitions,
            after_cursor=check.opt_inst_param(
                after_cursor, "after_cursor", (int, RunShardedEventsCursor)
            ),
            before_cursor=check.opt_inst_param(
                before_cursor, "before_cursor", (int, RunShardedEventsCursor)
            ),
            after_timestamp=check.opt_float_param(after_timestamp, "after_timestamp"),
            before_timestamp=check.opt_float_param(before_timestamp, "before_timestamp"),
            storage_ids=check.opt_nullable_sequence_param(storage_ids, "storage_ids", of_type=int),
        )

    @staticmethod
    def get_cursor_params(
        cursor: Optional[str] = None, ascending: bool = False
    ) -> tuple[Optional[int], Optional[int]]:
        if not cursor:
            return None, None

        cursor_obj = EventLogCursor.parse(cursor)
        if not cursor_obj.is_id_cursor():
            check.failed(f"Invalid cursor for fetching event records: {cursor}")
        after_cursor = cursor_obj.storage_id() if ascending else None
        before_cursor = cursor_obj.storage_id() if not ascending else None
        return before_cursor, after_cursor

    @property
    def tags(self) -> Optional[Mapping[str, Union[str, Sequence[str]]]]:
        return None


@whitelist_for_serdes
class AssetRecordsFilter(
    NamedTuple(
        "_AssetRecordsFilter",
        [
            ("asset_key", PublicAttr[AssetKey]),
            ("asset_partitions", PublicAttr[Optional[Sequence[str]]]),
            ("after_timestamp", PublicAttr[Optional[float]]),
            ("before_timestamp", PublicAttr[Optional[float]]),
            ("after_storage_id", PublicAttr[Optional[int]]),
            ("before_storage_id", PublicAttr[Optional[int]]),
            ("storage_ids", PublicAttr[Optional[Sequence[int]]]),
        ],
    )
):
    """Defines a set of filter fields for fetching a set of asset event records.

    Args:
        asset_key (Optional[AssetKey]): Asset key for which to get asset event entries / records.
        asset_partitions (Optional[List[str]]): Filter parameter such that only asset
            events with a partition value matching one of the provided values are returned.  Only
            valid when the `asset_key` parameter is provided.
        after_timestamp (Optional[float]): Filter parameter such that only event records for
            events with timestamp greater than the provided value are returned.
        before_timestamp (Optional[float]): Filter parameter such that only event records for
            events with timestamp less than the provided value are returned.
        after_storage_id (Optional[float]): Filter parameter such that only event records for
            events with storage_id greater than the provided value are returned.
        before_storage_id (Optional[float]): Filter parameter such that only event records for
            events with storage_id less than the provided value are returned.
        storage_ids (Optional[Sequence[int]]): Filter parameter such that only event records for
            the given storage ids are returned.
        tags (Optional[Mapping[str, Union[str, Sequence[str]]]]): Filter parameter such that only
            events with the given event tags are returned
    """

    def __new__(
        cls,
        asset_key: AssetKey,
        asset_partitions: Optional[Sequence[str]] = None,
        after_timestamp: Optional[float] = None,
        before_timestamp: Optional[float] = None,
        after_storage_id: Optional[int] = None,
        before_storage_id: Optional[int] = None,
        storage_ids: Optional[Sequence[int]] = None,
    ):
        return super().__new__(
            cls,
            asset_key=check.inst_param(asset_key, "asset_key", AssetKey),
            asset_partitions=check.opt_nullable_sequence_param(
                asset_partitions, "asset_partitions", of_type=str
            ),
            after_timestamp=check.opt_float_param(after_timestamp, "after_timestamp"),
            before_timestamp=check.opt_float_param(before_timestamp, "before_timestamp"),
            after_storage_id=check.opt_int_param(after_storage_id, "after_storage_id"),
            before_storage_id=check.opt_int_param(before_storage_id, "before_storage_id"),
            storage_ids=check.opt_nullable_sequence_param(storage_ids, "storage_ids", of_type=int),
        )

    def to_event_records_filter(
        self, event_type: AssetEventType, cursor: Optional[str] = None, ascending: bool = False
    ) -> EventRecordsFilter:
        before_cursor_storage_id, after_cursor_storage_id = EventRecordsFilter.get_cursor_params(
            cursor, ascending
        )
        if self.before_storage_id and before_cursor_storage_id:
            before_cursor = min(self.before_storage_id, before_cursor_storage_id)
        else:
            before_cursor = (
                before_cursor_storage_id if before_cursor_storage_id else self.before_storage_id
            )
        if self.after_storage_id and after_cursor_storage_id:
            after_cursor = max(self.after_storage_id, after_cursor_storage_id)
        else:
            after_cursor = (
                after_cursor_storage_id if after_cursor_storage_id else self.after_storage_id
            )

        return EventRecordsFilter(
            event_type=event_type,
            asset_key=self.asset_key,
            asset_partitions=self.asset_partitions,
            after_cursor=after_cursor,
            before_cursor=before_cursor,
            after_timestamp=self.after_timestamp,
            before_timestamp=self.before_timestamp,
            storage_ids=self.storage_ids,
        )

    @property
    def tags(self) -> Optional[Mapping[str, Union[str, Sequence[str]]]]:
        return None


@whitelist_for_serdes
class RunStatusChangeRecordsFilter(
    NamedTuple(
        "_RunStatusChangeRecordsFilter",
        [
            ("event_type", PublicAttr[RunStatusChangeEventType]),
            ("after_timestamp", PublicAttr[Optional[float]]),
            ("before_timestamp", PublicAttr[Optional[float]]),
            ("after_storage_id", PublicAttr[Optional[int]]),
            ("before_storage_id", PublicAttr[Optional[int]]),
            ("storage_ids", PublicAttr[Optional[Sequence[int]]]),
            ("job_names", Optional[Sequence[str]]),
        ],
    )
):
    """Defines a set of filter fields for fetching a set of asset event records.

    Args:
        event_type (DagsterEventType): Filter argument for dagster event type
        after_timestamp (Optional[float]): Filter parameter such that only event records for
            events with timestamp greater than the provided value are returned.
        before_timestamp (Optional[float]): Filter parameter such that only event records for
            events with timestamp less than the provided value are returned.
        after_storage_id (Optional[float]): Filter parameter such that only event records for
            events with storage_id greater than the provided value are returned.
        before_storage_id (Optional[float]): Filter parameter such that only event records for
            events with storage_id less than the provided value are returned.
        storage_ids (Optional[Sequence[int]]): Filter parameter such that only event records for
            the given storage ids are returned.
    """

    def __new__(
        cls,
        event_type: RunStatusChangeEventType,
        after_timestamp: Optional[float] = None,
        before_timestamp: Optional[float] = None,
        after_storage_id: Optional[int] = None,
        before_storage_id: Optional[int] = None,
        storage_ids: Optional[Sequence[int]] = None,
        job_names: Optional[Sequence[str]] = None,
    ):
        if event_type not in EVENT_TYPE_TO_PIPELINE_RUN_STATUS:
            check.failed("Invalid event type for run status change event filter")

        return super().__new__(
            cls,
            event_type=check.inst_param(event_type, "event_type", DagsterEventType),
            after_timestamp=check.opt_float_param(after_timestamp, "after_timestamp"),
            before_timestamp=check.opt_float_param(before_timestamp, "before_timestamp"),
            after_storage_id=check.opt_int_param(after_storage_id, "after_storage_id"),
            before_storage_id=check.opt_int_param(before_storage_id, "before_storage_id"),
            storage_ids=check.opt_nullable_sequence_param(storage_ids, "storage_ids", of_type=int),
            job_names=check.opt_nullable_sequence_param(job_names, "job_names", of_type=str),
        )

    def to_event_records_filter_without_job_names(
        self, cursor: Optional[str] = None, ascending: bool = False
    ) -> EventRecordsFilter:
        before_cursor_storage_id, after_cursor_storage_id = EventRecordsFilter.get_cursor_params(
            cursor, ascending
        )
        if self.before_storage_id and before_cursor_storage_id:
            before_cursor = min(self.before_storage_id, before_cursor_storage_id)
        else:
            before_cursor = (
                before_cursor_storage_id if before_cursor_storage_id else self.before_storage_id
            )
        if self.after_storage_id and after_cursor_storage_id:
            after_cursor = max(self.after_storage_id, after_cursor_storage_id)
        else:
            after_cursor = (
                after_cursor_storage_id if after_cursor_storage_id else self.after_storage_id
            )

        return EventRecordsFilter(
            event_type=self.event_type,
            after_cursor=after_cursor,
            before_cursor=before_cursor,
            after_timestamp=self.after_timestamp,
            before_timestamp=self.before_timestamp,
            storage_ids=self.storage_ids,
        )



from typing import Any, Literal, cast

from typing_extensions import TypeVar, get_args, get_origin

from dagster import _check as check
from dagster._core.definitions.metadata.metadata_set import NamespacedKVSet
from dagster._model.pydantic_compat_layer import model_fields
from dagster._utils.typing_api import is_closed_python_optional_type

T_NamespacedTagSet = TypeVar("T_NamespacedTagSet", bound="NamespacedTagSet")


class NamespacedTagSet(NamespacedKVSet):
    """Extend this class to define a set of tags in the same namespace.

    Supports splatting to a dictionary that can be placed inside a tags argument along with
    other tags.

    .. code-block:: python

        my_tags: NamespacedTagsSet = ...
        @asset(
            tags={**my_tags}
        )
        def my_asset():
            pass
    """

    def __init__(self, *args, **kwargs) -> None:
        for field_name, field in model_fields(self).items():
            annotation_type = field.annotation

            is_optional = is_closed_python_optional_type(annotation_type)
            is_optional_str = is_optional and str in get_args(annotation_type)
            is_optional_literal = (
                is_optional and get_origin(get_args(annotation_type)[0]) == Literal
            )
            if not (
                is_optional_str
                or annotation_type is str
                or is_optional_literal
                or annotation_type is Literal
            ):
                check.failed(
                    f"Type annotation for field '{field_name}' is not str, Optional[str], Literal, "
                    f"or Optional[Literal]. Is {annotation_type}."
                )
        super().__init__(*args, **kwargs)

    @classmethod
    def _extract_value(cls, field_name: str, value: Any) -> str:
        """Since all tag values are strings, we don't need to do any type coercion."""
        return cast(str, value)



import logging
import os
import sys
from collections.abc import Iterator, Mapping, Sequence
from typing import TYPE_CHECKING, Callable, Optional

import dagster._check as check
from dagster import DagsterEventType
from dagster._core.instance import DagsterInstance
from dagster._core.storage.dagster_run import RunRecord, RunsFilter
from dagster._core.workspace.context import IWorkspaceProcessContext
from dagster._daemon.auto_run_reexecution.auto_run_reexecution import (
    consume_new_runs_for_automatic_reexecution,
)
from dagster._daemon.daemon import IntervalDaemon
from dagster._daemon.utils import DaemonErrorCapture

if TYPE_CHECKING:
    from dagster._core.events.log import EventLogEntry

_INTERVAL_SECONDS = int(os.environ.get("DAGSTER_EVENT_LOG_CONSUMER_DAEMON_INTERVAL_SECONDS", 5))
_EVENT_LOG_FETCH_LIMIT = int(os.environ.get("DAGSTER_EVENT_LOG_CONSUMER_DAEMON_FETCH_LIMIT", 500))

DAGSTER_EVENT_TYPES = [DagsterEventType.RUN_FAILURE, DagsterEventType.RUN_SUCCESS]


class EventLogConsumerDaemon(IntervalDaemon):
    def __init__(
        self,
        interval_seconds: int = _INTERVAL_SECONDS,
        event_log_fetch_limit: int = _EVENT_LOG_FETCH_LIMIT,
    ):
        super().__init__(interval_seconds=interval_seconds)
        self._event_log_fetch_limit = event_log_fetch_limit

    @classmethod
    def daemon_type(cls) -> str:
        return "EVENT_LOG_CONSUMER"

    @property
    def handle_updated_runs_fns(
        self,
    ) -> Sequence[
        Callable[[IWorkspaceProcessContext, Sequence[RunRecord], logging.Logger], Iterator]
    ]:
        """List of functions that will be called with the list of run records that have new events."""
        return [consume_new_runs_for_automatic_reexecution]

    def run_iteration(self, workspace_process_context: IWorkspaceProcessContext):
        instance = workspace_process_context.instance
        # get the persisted cursor for each event type
        persisted_cursors = _fetch_persisted_cursors(instance, DAGSTER_EVENT_TYPES, self._logger)

        # Get the current greatest event id before we query for the specific event types
        overall_max_event_id = instance.event_log_storage.get_maximum_record_id()

        events: list[EventLogEntry] = []
        new_cursors: dict[
            DagsterEventType, int
        ] = {}  # keep these in memory until we handle the events
        for event_type in DAGSTER_EVENT_TYPES:
            yield

            cursor = persisted_cursors[event_type]
            if cursor is None:
                # if we don't have a cursor for this event type, start at the top of the event log and ignore older events. Otherwise enabling the daemon would result in retrying all old runs.
                cursor = overall_max_event_id or 0

            events_by_log_id_for_type = instance.event_log_storage.get_logs_for_all_runs_by_log_id(
                after_cursor=cursor,
                dagster_event_type={event_type},
                limit=self._event_log_fetch_limit,
            )

            events.extend(events_by_log_id_for_type.values())

            # calculate the new cursor for this event type
            new_cursors[event_type] = get_new_cursor(
                cursor,
                overall_max_event_id,
                self._event_log_fetch_limit,
                list(events_by_log_id_for_type.keys()),
            )

        if events:
            run_ids = list({event.run_id for event in events})
            run_records = instance.get_run_records(filters=RunsFilter(run_ids=run_ids))

            # call each handler with the list of runs that have events
            for fn in self.handle_updated_runs_fns:
                try:
                    yield from fn(workspace_process_context, run_records, self._logger)
                except Exception:
                    DaemonErrorCapture.process_exception(
                        sys.exc_info(),
                        logger=self._logger,
                        log_message=f"Error calling event event log consumer handler: {fn.__name__}",
                    )

        # persist cursors now that we've processed all the events through the handlers
        _persist_cursors(instance, new_cursors)


def _create_cursor_key(event_type: DagsterEventType) -> str:
    check.inst_param(event_type, "event_type", DagsterEventType)

    return f"EVENT_LOG_CONSUMER_CURSOR-{event_type.value}"


def _fetch_persisted_cursors(
    instance: DagsterInstance, event_types: Sequence[DagsterEventType], logger: logging.Logger
) -> dict[DagsterEventType, Optional[int]]:
    check.inst_param(instance, "instance", DagsterInstance)
    check.sequence_param(event_types, "event_types", of_type=DagsterEventType)

    # get the persisted cursor for each event type
    persisted_cursors = instance.daemon_cursor_storage.get_cursor_values(
        {_create_cursor_key(event_type) for event_type in event_types}
    )

    fetched_cursors: dict[DagsterEventType, Optional[int]] = {}
    for event_type in event_types:
        raw_cursor_value = persisted_cursors.get(_create_cursor_key(event_type))

        if raw_cursor_value is None:
            logger.warn(f"No cursor for event type {event_type}, ignoring older events")
            fetched_cursors[event_type] = None
        else:
            try:
                cursor_value = int(raw_cursor_value)
            except ValueError:
                logger.exception(f"Invalid cursor for event_type {event_type}: {raw_cursor_value}")
                raise
            fetched_cursors[event_type] = cursor_value

    return fetched_cursors


def _persist_cursors(instance: DagsterInstance, cursors: Mapping[DagsterEventType, int]) -> None:
    check.inst_param(instance, "instance", DagsterInstance)
    check.mapping_param(cursors, "cursors", key_type=DagsterEventType, value_type=int)

    if cursors:
        instance.daemon_cursor_storage.set_cursor_values(
            {
                _create_cursor_key(event_type): str(cursor_value)
                for event_type, cursor_value in cursors.items()
            }
        )


def get_new_cursor(
    persisted_cursor: int,
    overall_max_event_id: Optional[int],
    fetch_limit: int,
    new_event_ids: Sequence[int],
) -> int:
    """Return the new cursor value for an event type, or None if one shouldn't be persisted.

    The cursor is guaranteed to be:

    - greater than or equal to any id in new_event_ids (otherwise we could process an event twice)
    - less than the id of any event of the desired type that hasn't been fetched yet (otherwise we
      could skip events)

    This method optimizes for moving the cursor as far forward as possible, using
    overall_max_event_id.
    """
    check.int_param(persisted_cursor, "persisted_cursor")
    check.opt_int_param(overall_max_event_id, "overall_max_event_id")
    check.int_param(fetch_limit, "fetch_limit")
    check.sequence_param(new_event_ids, "new_event_ids", of_type=int)

    if overall_max_event_id is None:
        # We only get here if the event log was empty when we queried it for the overall max.
        if new_event_ids:
            # We only get here if some events snuck in after the max id query. Set the cursor
            # to the max event id of the new events.
            return max(new_event_ids)

        # Event log is empty, set the cursor to 0 so we pick up the next events. Otherwise we'd skip to
        # the latest event if no cursor was set.
        return 0

    if not new_event_ids:
        # No new events, so we can skip to overall_max_event_id because we queried that first, so we
        # know there are no relevant events up to that id.
        return overall_max_event_id

    # these should be ordered, but we won't assume
    max_new_event_id = max(new_event_ids)

    check.invariant(
        max_new_event_id > persisted_cursor,
        f"The new cursor {max_new_event_id} should be greater than the previous {persisted_cursor}",
    )

    num_new_events = len(new_event_ids)
    check.invariant(
        num_new_events <= fetch_limit,
        "Query returned more than the limit!",
    )

    if num_new_events == fetch_limit:
        # We got back the limit number of events, so the only thing we can do is move the cursor
        # forward to the max event id of the new events. It's possible for the very next log id
        # to be the desired type. There's no way to skip ahead.
        return max_new_event_id
    else:
        # We got back fewer than the limit number of events, so we may be able to skip ahead. Since
        # we queried for overall_max_event_id before we queried for events of the desired type, we
        # know that there can be no events of the desired type with ids less than overall_max_event_id
        # that we haven't fetched yet (they would have been in this query, up until it reached the
        # fetch limit). Thus we can skip ahead to overall_max_event_id.
        if overall_max_event_id >= max_new_event_id:
            return overall_max_event_id

        # There's also a rare case where more events of our desired type snuck in after the query
        # for overall_max_event_id but before the specific event query. In this case, we just move
        # the cursor forward to the max event id of the new events.
        return max_new_event_id
